from dataclasses import dataclass
from enum import auto
from typing import Any

from mstrio.api import prompts
from mstrio.connection import Connection
from mstrio.helpers import VersionException
from mstrio.modeling import Attribute
from mstrio.modeling.expression import ExpressionFormat
from mstrio.modeling.schema.helpers import ObjectSubType
from mstrio.object_management import (
    Folder,
    SearchPattern,
    full_search,
    get_folder_id_from_path,
)
from mstrio.server import Project
from mstrio.types import ObjectSubTypes, ObjectTypes
from mstrio.users_and_groups.user import User
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity, MoveMixin
from mstrio.utils.enum_helper import AutoName, AutoUpperName, get_enum_val
from mstrio.utils.helper import (
    Dictable,
    filter_params_for_func,
    find_object_with_name,
    get_owner_id,
    is_valid_str_id,
)
from mstrio.utils.resolvers import (
    get_project_id_from_params_set,
    validate_owner_key_in_filters,
)
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.version_helper import is_server_min_version, method_version_handler


class PromptType(AutoUpperName):
    """Enum for prompt types."""

    UNSUPPORTED = auto()
    VALUE = auto()
    ELEMENTS = auto()
    EXPRESSION = auto()
    OBJECTS = auto()
    LEVEL = auto()


class PersonalAnswerMode(AutoName):
    """Enum for personal answer modes in prompt restrictions."""

    NONE = auto()
    SINGLE = auto()
    MULTIPLE = auto()


@dataclass
class PromptRestrictions(Dictable):
    """Class representing prompt restrictions."""

    allow_personal_answers: PersonalAnswerMode | str
    required: bool
    max_elements_per_selection: int | None = None
    min: int | None = None
    max: int | None = None

    def __post_init__(self):
        """Convert string values to enum if needed."""
        if isinstance(self.allow_personal_answers, str):
            if self.allow_personal_answers.lower() not in (
                possible_values := [mode.value for mode in PersonalAnswerMode]
            ):
                raise ValueError(
                    f"Invalid value '{self.allow_personal_answers}' for "
                    f"allow_personal_answers. Possible values are: "
                    f"{possible_values}"
                )
            self.allow_personal_answers = PersonalAnswerMode(
                self.allow_personal_answers.lower()
            )


@method_version_handler('11.4.0600')
def list_prompts(
    connection: Connection,
    name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    project: 'Project | str | None' = None,
    project_id: str | None = None,
    project_name: str | None = None,
    search_pattern: SearchPattern | int = SearchPattern.CONTAINS,
    show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
    **filters,
) -> list['Prompt'] | list[dict]:
    """Get list of Prompt objects or dicts with them.

    Args:
        connection (object): Strategy One connection object returned by
            `connection.Connection()`
        name (optional, str): value the search pattern is set to, which
            will be applied to the names of prompts being searched
        to_dictionary (optional, bool): If `True` returns dictionaries, by
            default (`False`) returns `Prompt` objects.
        limit (optional, int): limit the number of elements returned. If `None`
            (default), all objects are returned.
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        search_pattern (SearchPattern enum or int, optional): pattern to
            search for, such as Begin With or Exactly. Possible values are
            available in ENUM `mstrio.object_management.SearchPattern`.
            Default value is CONTAINS (4).
        show_expression_as (optional, enum or str): specify how expressions
            should be presented.
            Available values:
                - `ExpressionFormat.TREE` or `tree` (default)
                - `ExpressionFormat.TOKENS or `tokens`
        **filters: Available filter parameters: ['id', 'subtype',
            'date_created', 'date_modified', 'version', 'acg', 'owner',
            'ext_type']

    Returns:
        list of prompt objects or list of prompt dictionaries.
    """

    proj_id = get_project_id_from_params_set(
        connection,
        project,
        project_id,
        project_name,
    )

    validate_owner_key_in_filters(filters)

    objects = full_search(
        connection,
        object_types=ObjectTypes.PROMPT,
        project=proj_id,
        name=name,
        pattern=search_pattern,
        limit=limit,
        **filters,
    )
    if to_dictionary:
        return objects
    else:
        show_expression_as = (
            show_expression_as
            if isinstance(show_expression_as, ExpressionFormat)
            else ExpressionFormat(show_expression_as)
        )
        return [
            Prompt.from_dict(
                source={**obj, 'show_expression_as': show_expression_as},
                connection=connection,
                with_missing_value=False,
            )
            for obj in objects
        ]


class Prompt(CopyMixin, DeleteMixin, Entity, MoveMixin):
    """A Strategy One class representing a prompt.

    This class can be used in two modes:
        1. Answer mode: Create local prompt objects for providing answers
            without API calls
        2. API mode: Fetch and manage prompt objects from the MicroStrategy
            server

    Attributes:
        id (str): ID of the prompt
        name (str): Name of the prompt
        prompt_type (str): Type of the prompt. Possible values are:
            - UNSUPPORTED
            - VALUE
            - ELEMENTS
            - EXPRESSION
            - OBJECTS
            - LEVEL
        sub_type (ObjectSubType): Sub-type of the prompt (e.g., PROMPT_STRING,
            PROMPT_DOUBLE)
        title (str): Title of the prompt displayed in the interface
        instruction (str): Instruction text for the prompt
        default_answer (dict): Default answer configuration
        restriction (PromptRestrictions): Restrictions applied to the prompt
        answers (Any | list[Any]): Answer(s) to the prompt (answer mode only)
        key (str): Unique key of the prompt (answer mode only)
        use_default (bool): Whether to use default value (answer mode only)
        date_created (str): Date when the prompt was created
        date_modified (str): Date when the prompt was last modified
        owner (User): Owner of the prompt
        personal_answers (list[dict]): List of personal answers for the prompt

    Examples:
        Creating a prompt for answering (answer mode):
        >>> prompt = Prompt(
        >>>     key='8891A8AF4A747A3EA8506DBB6189F252@0@10',
        >>>     answers=[{'name': 'Cost', 'id':
                        '7FD5B69611D5AC76C000D98A4CC5F24F'}],
        >>>     type='OBJECTS'
        >>> )

        Fetching a prompt from server (API mode):
        >>> prompt = Prompt(connection, id='prompt_id')
        >>> prompt = Prompt(connection, name='My Prompt')

        Creating a new value prompt:
        >>> prompt = Prompt.create_value_prompt(
        >>>     connection=conn,
        >>>     name='Number Input',
        >>>     sub_type=ObjectSubType.PROMPT_DOUBLE,
        >>>     destination_folder=folder_id
        >>> )

    Note:
        In answer mode, only key needs to be provided along with answers or
            use_default.
        In API mode, either id or name must be provided along with a connection
            object.
    """

    _OBJECT_TYPE = ObjectTypes.PROMPT

    _API_GETTERS = {
        **Entity._API_GETTERS,
        (
            'id',
            'name',
            'sub_type',
            'title',
            'instruction',
            'default_answer',
            'restriction',
            'date_created',
            'date_modified',
        ): prompts.get_prompt,
    }
    _API_PATCH = {
        ('folder_id', 'comments', 'hidden', 'owner'): (
            objects_processors.update,
            'partial_put',
        ),
        (
            'title',
            'instruction',
            'default_answer',
            'restriction',
            'information',
            'name',
            'sub_type',
        ): (prompts.update_prompt, 'put'),
    }
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'owner': User.from_dict,
        'sub_type': ObjectSubType,
        'restriction': PromptRestrictions.from_dict,
    }
    _PATCH_PATH_TYPES = {
        **Entity._PATCH_PATH_TYPES,
        'title': str,
        'instruction': str,
        'default_answer': dict,
        'restriction': dict,
    }

    def __init__(
        self,
        connection: Connection | None = None,
        id: str | None = None,
        key: str | None = None,
        name: str | None = None,
        type: 'str | PromptType | None' = None,
        answers: Any | list[Any] | None = None,
        use_default: bool = False,
    ) -> None:
        """Initialize a new instance of Prompt class.

        Can be used in two ways:
            1. Answer mode: Provide key and answers to create a local prompt
                object without API calls
            2. API mode: Provide connection and identification to fetch from
                server

        Args:
            connection (Connection, optional): Strategy One connection object
            id (str, optional): Identifier of a prompt object
            key (str, optional): Unique key of the prompt
            name (str, optional): Name of a prompt object
            type (str, PromptType, optional): Type of the prompt (ELEMENTS,
                OBJECTS, etc.)
            answers (Any | list[Any], optional): Answers to the prompt
            use_default (bool, optional): Whether to use default value.
                Defaults to False.
        """
        # Check if this is answer mode - key and answers provided
        if (
            connection is None
            and key is not None
            and (answers is not None or use_default)
        ):
            # Create local prompt object without API call
            self._init_local_prompt(
                type=type,
                answers=answers,
                key=key,
                id=id,
                name=name,
                use_default=use_default,
            )
        elif connection is not None and (id is not None or name is not None):
            # API mode - fetch from server
            if not is_server_min_version(connection, '11.4.0600'):
                raise VersionException(
                    'Initializing Prompt in API mode requires I-Server version '
                    '11.4.0600 or later.'
                )
            if id is None:
                prompt = find_object_with_name(
                    connection=connection,
                    cls=self.__class__,
                    name=name,
                    listing_function=list_prompts,
                    search_pattern=SearchPattern.EXACTLY,
                )
                id = prompt['id']

            super().__init__(
                connection=connection, object_id=id, project_id=connection.project_id
            )
        else:
            raise ValueError(
                "Either provide (key + answers/use_default) without connection for "
                "answer mode or (connection + id/name) for API-based initialization"
            )

    def _init_local_prompt(
        self,
        type: str | None = None,
        answers: Any | list[Any] | None = None,
        key: str | None = None,
        id: str | None = None,
        name: str | None = None,
        use_default: bool = False,
    ):
        """Initialize prompt with local data (answers mode)."""
        self._connection = None
        self._id = id
        self.prompt_type = type
        self.answers = answers
        self.key = key
        self.name = name
        self.use_default = use_default

    def _init_variables(self, default_value, **kwargs) -> None:
        """Initialize all properties of the Prompt class."""
        super()._init_variables(default_value=default_value, **kwargs)
        self.answers = kwargs.get('answers')
        self.use_default = kwargs.get('use_default')
        self.name = kwargs.get('name')
        self.default_answer = kwargs.get('default_answer', {})
        self.required = kwargs.get('required')
        self.closed = kwargs.get('closed')
        self.instruction = kwargs.get('instruction')
        self.question = kwargs.get('question', {})
        self.sub_type = (
            ObjectSubType(kwargs.get('sub_type')) if kwargs.get('sub_type') else None
        )
        self._restriction = (
            PromptRestrictions.from_dict(kwargs['restriction'])
            if kwargs.get('restriction')
            else None
        )

        self.data_type = kwargs.get('data_type')
        self.title = kwargs.get('title')
        self._personal_answers = kwargs.get('personal_answers', [])
        self.prompt_type = (
            PromptType(kwargs.get('type'))
            if kwargs.get('type') in PromptType.__members__
            else None
        )
        self.key = kwargs.get('key')

    @classmethod
    @method_version_handler('11.4.0600')
    def create(
        cls,
        connection: Connection,
        prompt_data: dict,
        name: str,
        sub_type: ObjectSubTypes | str,
        destination_folder: Folder | str,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
    ) -> 'Prompt':
        """Create a new prompt object.

        Args:
            connection (Connection): Strategy One connection object.
            prompt_data (dict): Dictionary containing prompt configuration data.
            name (str): Name of the prompt.
            sub_type (ObjectSubType | str): Sub-type of the prompt.
            destination_folder (Folder | str): Destination folder object or
                folder ID/path.
            show_expression_as (ExpressionFormat | str, optional): How
                expressions should be presented. Defaults to
                ExpressionFormat.TREE.

        Returns:
            Prompt: An instance of the Prompt class.
        """
        if isinstance(destination_folder, Folder):
            destination_folder_id = destination_folder.id
        else:
            if '/' in destination_folder:
                destination_folder_id = get_folder_id_from_path(
                    connection=connection, folder_path=destination_folder
                )
            else:
                destination_folder_id = destination_folder

        if 'information' not in prompt_data:
            prompt_data['information'] = {}

        prompt_data['information']['name'] = name
        prompt_data['information']['destinationFolderId'] = destination_folder_id
        prompt_data['information']['subType'] = (
            sub_type.value if isinstance(sub_type, ObjectSubType) else sub_type
        )
        response = prompts.create_prompt(
            connection,
            prompt_data,
            show_expression_as=get_enum_val(show_expression_as, ExpressionFormat),
        ).json()
        return cls.from_dict(source=response, connection=connection)

    @classmethod
    @method_version_handler('11.4.0600')
    def create_value_prompt(
        cls,
        connection: Connection,
        name: str,
        sub_type: ObjectSubType | str,
        destination_folder: Folder | str,
        title: str | None = None,
        instruction: str | None = None,
        restrictions: dict | PromptRestrictions | None = None,
        default_answer: dict | str | int | float | None = None,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
    ) -> 'Prompt':
        """Create a new value prompt object.

        Args:
            connection (Connection): Strategy One connection object.
            name (str): Name of the prompt.
            sub_type (ObjectSubType | str): Sub-type of the prompt
                (PROMPT_STRING, PROMPT_DATE, PROMPT_DOUBLE, PROMPT_BIG_DECIMAL).
            destination_folder (Folder | str): Destination folder object or
                folder ID/path.
            title (str, optional): Title of the prompt. If not provided, will be
                auto-generated based on sub_type.
            instruction (str, optional): Instruction for the prompt. If not
                provided, will be auto-generated.
            restrictions (dict | PromptRestrictions, optional): Restrictions for
                the prompt.
            default_answer (dict | str | int | float, optional): Default answer
                value. Can be a dictionary, string, integer, or floating-point
                number.
            show_expression_as (ExpressionFormat | str, optional): How
            expressions should be presented. Defaults to ExpressionFormat.TREE.

        Returns:
            Prompt: An instance of the Prompt class.
        """
        VALUE_PROMPT_SUB_TYPES = [
            ObjectSubType.PROMPT_STRING,
            ObjectSubType.PROMPT_DATE,
            ObjectSubType.PROMPT_DOUBLE,
            ObjectSubType.PROMPT_BIG_DECIMAL,
        ]

        if not isinstance(sub_type, ObjectSubType):
            sub_type = ObjectSubType(sub_type)
        if sub_type not in VALUE_PROMPT_SUB_TYPES:
            raise ValueError(
                f"Invalid sub_type '{sub_type}'. "
                f"Possible values are: {VALUE_PROMPT_SUB_TYPES}"
            )
        if not title:
            title_mapping = {
                ObjectSubType.PROMPT_STRING: 'Text',
                ObjectSubType.PROMPT_DATE: 'Date',
                ObjectSubType.PROMPT_DOUBLE: 'Number',
                ObjectSubType.PROMPT_BIG_DECIMAL: 'Big Decimal',
            }
            title = title_mapping.get(sub_type)
        if not instruction:
            instruction = f"Enter a value ({title})."
        if not restrictions:
            restrictions = PromptRestrictions(
                required=False,
                allow_personal_answers=PersonalAnswerMode.NONE.value,
            )
        if default_answer:
            if isinstance(default_answer, (str, int, float)):
                default_answer = {'value': default_answer}
        else:
            default_answer = {}

        prompt_data = cls.prepare_prompt_data(
            title=title,
            instruction=instruction,
            restrictions=restrictions,
            default_answer=default_answer,
            question={},
        )
        return cls.create(
            connection=connection,
            name=name,
            prompt_data=prompt_data,
            sub_type=sub_type,
            destination_folder=destination_folder,
            show_expression_as=show_expression_as,
        )

    @classmethod
    @method_version_handler('11.4.0600')
    def create_attr_elements_prompt(
        cls,
        connection: Connection,
        name: str,
        destination_folder: Folder | str,
        attribute: Attribute | str,
        title: str | None = None,
        instruction: str | None = None,
        restrictions: dict | PromptRestrictions | None = None,
        default_answer: dict | list | None = None,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
    ) -> 'Prompt':
        """Create a new attribute element prompt.

        Args:
            connection (Connection): Connection object.
            name (str): Name of the prompt.
            destination_folder (Folder | str): Destination folder object or
                folder ID/path.
            show_expression_as (ExpressionFormat | str, optional): How
                expressions should be presented. Defaults to
                ExpressionFormat.TREE.

        Returns:
            Prompt: An instance of the Prompt class.
        """

        if not isinstance(attribute, Attribute):
            if is_valid_str_id(attribute):
                attribute = Attribute(connection=connection, id=attribute)
            else:
                attribute = Attribute(connection=connection, name=attribute)
        if isinstance(attribute, Attribute):
            attribute = {
                'objectId': attribute.id,
                'subType': ObjectSubType.ATTRIBUTE.value,
                'name': attribute.name,
            }
        if not title:
            title = attribute.get('name', 'Attribute')
        if not instruction:
            instruction = f"Choose elements of {title}."
        if not restrictions:
            restrictions = PromptRestrictions(
                required=False,
                allow_personal_answers=PersonalAnswerMode.NONE.value,
            )
        if default_answer:
            if isinstance(default_answer, list):
                default_answer = {'elements': default_answer}
        else:
            default_answer = {}

        prompt_data = cls.prepare_prompt_data(
            title=title,
            instruction=instruction,
            restrictions=restrictions,
            default_answer=default_answer,
            question={'attribute': attribute, 'listAllElements': True},
        )
        return cls.create(
            connection=connection,
            name=name,
            prompt_data=prompt_data,
            sub_type=ObjectSubType.PROMPT_ELEMENTS,
            destination_folder=destination_folder,
            show_expression_as=show_expression_as,
        )

    @staticmethod
    def prepare_prompt_data(
        title, instruction, restrictions, default_answer, question
    ) -> dict:
        return {
            'title': title,
            'instruction': instruction,
            'restriction': (
                restrictions.to_dict()  # NOSONAR due to SonarQube issue
                # https://community.sonarsource.com/t/python-s930-fp-on-pathlib-path-method-calls/134261/7
                if isinstance(restrictions, PromptRestrictions)
                else restrictions
            ),
            'defaultAnswer': default_answer,
            'question': question,
        }

    @method_version_handler('11.4.0600')
    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        default_answer: dict | None = None,
        instruction: str | None = None,
        restriction: dict | PromptRestrictions | None = None,
        title: str | None = None,
        owner: str | User | None = None,
        owner_id: str | None = None,
        owner_username: str | None = None,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
    ) -> None:
        """Alter prompt properties.

        Args:
        name (str, optional): Name of the prompt.
        description (str, optional): Description of the prompt.
        default_answer (dict , optional): Default answer value.
        instruction (str, optional): Instruction text for the prompt.
        restriction (dict | PromptRestrictions, optional): Restrictions for the
            prompt. Can be a dictionary or PromptRestrictions instance.
        title (str, optional): Title of the prompt.
        owner (str | User, optional): New owner of the prompt.
            Can be a User object, user ID or username.
        owner_id (str, optional): ID of the new owner.
        owner_username (str, optional): Username of the new owner.
        show_expression_as (ExpressionFormat | str, optional): How expressions
        should be presented. Defaults to ExpressionFormat.TREE.
        """
        if owner or owner_id or owner_username:
            owner = get_owner_id(self.connection, owner, owner_id, owner_username)
            owner_id = None
            owner_username = None
        properties = filter_params_for_func(self.alter, locals(), exclude=['self'])
        properties['subType'] = self.sub_type.value
        self._alter_properties(**properties)

    @method_version_handler('11.4.0600')
    def add_personal_answer(
        self, answer: Any, name_of_personal_answer: str, set_as_default: bool = False
    ) -> bool:
        """Add a personal answer to the prompt.

        Args:
            answer (Any): The personal answer to be added.
            name_of_personal_answer (str): The name of the personal answer.
            set_as_default (bool): Whether to set the answer as the default.

        Returns:
            bool: True if the personal answer was added successfully,
                False otherwise.
        """
        body = {
            "personalPromptAnswerName": name_of_personal_answer,
            "isDefaultPromptAnswer": set_as_default,
            "values": [answer],
        }
        response = prompts.create_personal_answer(
            connection=self.connection,
            id=self.id,
            project_id=self.connection.project_id,
            body=body,
        )
        if response.ok:
            new_answer = response.json()
            self._personal_answers.append(new_answer)
            return True
        else:
            return False

    def to_dict(self, camel_case: bool = True) -> dict:
        result = super().to_dict(camel_case)
        for key in ['promptType', 'prompt_type']:
            if value := result.pop(key, None):
                result['type'] = value.value if isinstance(value, PromptType) else value
                break
        return result

    @property
    def personal_answers(self) -> list[dict]:
        """Get personal answers for the prompt."""
        if not hasattr(self, '_personal_answers') or self._personal_answers is None:
            return []
        return self._personal_answers

    @property
    def restriction(self) -> PromptRestrictions | None:
        """Get the restriction for the prompt."""
        if not hasattr(self, '_restriction') or self._restriction is None:
            self.fetch('restriction')
        return self._restriction



"""
Disable all active agents in a project.

1. Provide name of the project to connect to
2. Connect to the environment using data from workstation
3. Get a list of all agents in the project
4. Disable every active agent on the list
"""

from mstrio.connection import get_connection
from mstrio.project_objects.agents import list_agents

# Define project to connect to
PROJECT_NAME = 'MicroStrategy Tutorial'

# Connect to environment without providing user credentials
# Variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Get a list of agents in the project
agents = list_agents(connection=conn)

# Disable every active agent
for agent in agents:
    if agent.status == 'enabled':
        agent.disable()



import json
from typing import TYPE_CHECKING, Union

from mstrio.utils.error_handlers import ErrorHandler
from mstrio.utils.helper import delete_none_values

if TYPE_CHECKING:
    from mstrio.connection import Connection
    from mstrio.object_management import CertifiedStatus, SearchPattern
    from mstrio.types import ObjectSubTypes, TypeOrSubtype
    from mstrio.utils.sessions import FuturesSessionWithRenewal


@ErrorHandler(err_msg="Error searching metadata.")
def store_search_instance(
    connection: 'Connection',
    project_id: str | None = None,
    name: str | None = None,
    pattern: int | None = None,
    domain: int | None = None,
    root: str | None = None,
    object_types: list[int] | None = None,
    uses_object: str | None = None,
    uses_recursive: bool = False,
    uses_one_of: bool | None = None,
    used_by_object: str | None = None,
    used_by_recursive: bool | None = None,
    used_by_one_of: bool | None = None,
    begin_modification_time: str | None = None,
    end_modification_time: str | None = None,
    error_msg: str | None = None,
):
    """
    Search the metadata and store an instance of search results.

    Args:
        connection(object): Strategy One connection object returned by
            `connection.Connection()`.
        project_id(string, optional): Project ID
        name(string, optional): Value the search pattern is set to, which will
            be applied to the names of object types being searched. For example,
            search for all report objects (type) whose name begins with
            (pattern) B (name).
        pattern(integer or enum class object, optional): Pattern to search for,
            such as Begin With or Exactly. Possible values are available in
            ENUM mstrio.object_management.SearchPattern.
            Default value is CONTAINS (4).
        domain(integer or enum class object, optional): Domain where the search
            will be performed, such as Local or Project. Possible values are
            available in ENUM mstrio.object_management.SearchDomain.
            Default value is DOMAIN_PROJECT (2).
        root(string, optional): Folder ID of the root folder where the search
            will be performed.
        object_types(list of enum class objects or integers, optional):
            Type(s) of object(s) to be searched, such as
            Folder, Attribute or User. Possible values available in ENUMs
            mstrio.types.ObjectTypes and
            mstrio.types.ObjectSubTypes.
        uses_object(string, optional): Constrain the search to only return
            objects which use the given object. The value should be 'objectId;
            object type', for example 'E02FE6DC430378A8BBD315AA791FC580;3'. It
            is not allowed to use both 'uses_object' and 'used_by_object' in one
            request.
        uses_recursive(boolean, optional): Control the Intelligence server to
            also find objects that use the given objects indirectly. Default
            value is false.
        uses_one_of(boolean): Control the Intelligence server to also find
            objects that use one of or all of given objects indirectly.
            Default value is false.
        used_by_object(string, optional): Constrain the search to only return
            objects which are used by the given object. The value should be
            'object Id; object type', for example:
            'E02FE6DC430378A8BBD315AA791FC580;3'. It is not allowed to use both
            'uses_object' and 'used_by_object' in one request.
        used_by_recursive(boolean, optional): Control the Intelligence server
            to also find objects that are used by the given objects indirectly.
            Default value is false.
        used_by_one_of(boolean): Control the Intelligence server to also
            find objects that are used by one of or all of given objects
            indirectly. Default value is false.
        begin_modification_time(string, optional): Field to filter request
            to return records newer than a given date in
            format 'yyyy-MM-dd'T'HH:mm:ssZ', for example 2021-04-04T06:33:32Z.
        end_modification_time(string, optional): Field to filter request
            to return records  older than a given date in
            format 'yyyy-MM-dd'T'HH:mm:ssZ', for example 2022-04-04T06:33:32Z.
        error_msg(string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response returned by the Strategy One REST server.
    """
    return connection.post(
        endpoint='/api/metadataSearches/results',
        headers={'X-MSTR-ProjectID': project_id},
        params={
            'name': name,
            'pattern': pattern,
            'domain': domain,
            'root': root,
            'type': object_types,
            'usesObject': uses_object,
            'usesRecursive': uses_recursive,
            'usedByObject': used_by_object,
            'usedByRecursive': used_by_recursive,
            'usesOneOf': uses_one_of,
            'usedByOneOf': used_by_one_of,
            'beginModificationTime': begin_modification_time,
            'endModificationTime': end_modification_time,
        },
    )


@ErrorHandler(err_msg="Error searching metadata.")
def store_search_instance_v2(
    connection: 'Connection',
    body: dict,
    project_id: str | None = None,
    name: str | None = None,
    pattern: int | None = None,
    domain: int | None = None,
    scope: str | None = None,
    root: str | None = None,
    object_types: list[int] | None = None,
    uses_object: str | None = None,
    uses_recursive: bool = False,
    uses_one_of: bool | None = None,
    used_by_object: str | None = None,
    used_by_recursive: bool | None = None,
    used_by_one_of: bool | None = None,
    begin_modification_time: str | None = None,
    end_modification_time: str | None = None,
    visibility: str | None = None,
    error_msg: str | None = None,
):
    """
    Search the metadata and store an instance of search results.

    Args:
        connection (object): Strategy One connection object returned by
            `connection.Connection()`.
        body (dict): Request body containing extra search params
        project_id (string, optional): Project ID
        name (string, optional): Value the search pattern is set to, which will
            be applied to the names of object types being searched. For example,
            search for all report objects (type) whose name begins with
            (pattern) B (name).
        pattern (integer, optional): Pattern to search for, such as Begin With
            or Exactly. Possible values are available in enum
            `mstrio.object_management.SearchPattern.` Default value is
            CONTAINS (4).
        domain (integer, optional): Domain where the search will be performed,
            such as Local or Project. Possible values are available in enum
            `mstrio.object_management.SearchDomain.` Default value is
            DOMAIN_PROJECT (2).
        scope (string, optional): Scope of the search in relation to managed
            objects. Possible values are available in enum
            `mstrio.object_management.SearchScope.` Default value is 'managed'.
        root (string, optional): Folder ID of the root folder where the search
            will be performed.
        object_types (list of integers, optional): Type(s) of object(s) to be
            searched. Possible values are available in enums
            `mstrio.types.ObjectTypes` and `mstrio.types.ObjectSubTypes`.
        uses_object (string, optional): Constrain the search to only return
            objects which use the given object. The value should be 'objectId;
            object type', for example 'E02FE6DC430378A8BBD315AA791FC580;3'. It
            is not allowed to use both 'uses_object' and 'used_by_object' in one
            request.
        uses_recursive (boolean, optional): Control the Intelligence server to
            also find objects that use the given objects indirectly. Default
            value is false.
        uses_one_of (boolean): Control the Intelligence server to also find
            objects that use one of or all of given objects indirectly.
            Default value is false.
        used_by_object (string, optional): Constrain the search to only return
            objects which are used by the given object. The value should be
            'object Id; object type', for example:
            'E02FE6DC430378A8BBD315AA791FC580;3'. It is not allowed to use both
            'uses_object' and 'used_by_object' in one request.
        used_by_recursive (boolean, optional): Control the Intelligence server
            to also find objects that are used by the given objects indirectly.
            Default value is false.
        used_by_one_of (boolean): Control the Intelligence server to also
            find objects that are used by one of or all of given objects
            indirectly. Default value is false.
        begin_modification_time (string, optional): Field to filter request
            to return records newer than a given date in
            format 'yyyy-MM-dd'T'HH:mm:ssZ', for example 2021-04-04T06:33:32Z.
        end_modification_time (string, optional): Field to filter request
            to return records  older than a given date in
            format 'yyyy-MM-dd'T'HH:mm:ssZ', for example 2022-04-04T06:33:32Z.
        visibility (string, optional): Filter the result based on the `hidden`
            field of objects. If not passed, no filtering will be applied.
            Available values are 'visible' and 'all'.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response returned by the Strategy One REST server.
    """
    return connection.post(
        endpoint='/api/v2/metadataSearches/results',
        headers={'X-MSTR-ProjectID': project_id},
        params={
            'name': name,
            'pattern': pattern,
            'domain': domain,
            'scope': scope,
            'root': root,
            'type': object_types,
            'usesObject': uses_object,
            'usesRecursive': uses_recursive,
            'usedByObject': used_by_object,
            'usedByRecursive': used_by_recursive,
            'usesOneOf': uses_one_of,
            'usedByOneOf': used_by_one_of,
            'beginModificationTime': begin_modification_time,
            'endModificationTime': end_modification_time,
            'visibility': visibility,
        },
        json=delete_none_values(body, recursion=True),
    )


@ErrorHandler(err_msg="Error getting search result for search with ID {search_id}")
def get_search_results(
    connection: 'Connection',
    search_id: str,
    project_id: str | None = None,
    offset: int = 0,
    limit: int = -1,
    error_msg: str | None = None,
):
    """
    Get search results in a list format.

    Args:
        connection(object): Strategy One connection object returned by
            `connection.Connection()`.
        search_id(string): Search ID (identifies the results for a previous
            search stored in I-Server memory)
        project_id(string, optional): Project ID
        offset(integer, optional): Starting point within the collection of
            returned results. Used to control paging behavior. Default value
            is 0.
        limit(integer, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default value is -1.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response returned by the Strategy One REST server.
    """
    return connection.get(
        endpoint='/api/metadataSearches/results',
        headers={'X-MSTR-ProjectID': project_id},
        params={'searchId': search_id, 'offset': offset, 'limit': limit},
    )


def get_search_results_async(
    future_session: 'FuturesSessionWithRenewal',
    search_id: str,
    project_id: str | None = None,
    offset: int = 0,
    limit: int = -1,
):
    """Get search results in a list format asynchronously.

    Args:
        future_session(object): Future Session object to call Strategy One REST
            Server asynchronously
        search_id(string): Search ID (identifies the results for a previous
            search stored in I-Server memory)
        project_id(string, optional): Project ID
        offset(integer, optional): Starting point within the collection of
            returned results. Used to control paging behavior. Default value
            is 0.
        limit(integer, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default value is -1.

    Returns:
        Future with HTTP response returned by the Strategy One REST server as
        a result.
    """
    endpoint = '/api/objects'
    headers = {'X-MSTR-ProjectID': project_id}
    params = {'searchId': search_id, 'offset': offset, 'limit': limit}
    future = future_session.get(endpoint=endpoint, headers=headers, params=params)
    return future


@ErrorHandler(
    err_msg="Error getting search result in atree format for search with ID {search_id}"
)
def get_search_results_tree_format(
    connection: 'Connection',
    search_id: str,
    project_id: str | None = None,
    offset: int = 0,
    limit: int = -1,
    error_msg: str | None = None,
):
    """
    Get search results in a tree format.

    Args:
        connection(object): Strategy One connection object returned by
            `connection.Connection()`.
        search_id(string): Search ID (identifies the results for a previous
            search stored in I-Server memory)
        project_id(string, optional): Project ID
        offset(integer, optional): Starting point within the collection of
            returned results. Used to control paging behavior. Default value
            is 0.
        limit(integer, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default value is -1.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response returned by the Strategy One REST server.
    """
    return connection.get(
        endpoint='/api/metadataSearches/results/tree',
        headers={'X-MSTR-ProjectID': project_id},
        params={'searchId': search_id, 'offset': offset, 'limit': limit},
    )


@ErrorHandler(err_msg="Error getting quick search result.")
def get_quick_search_result(
    connection,
    project_id: str | None = None,
    name: str | None = None,
    root: str | None = None,
    object_types: 'TypeOrSubtype | None' = None,
    pattern: Union['SearchPattern', int] | None = None,
    certified_status: 'CertifiedStatus | None' = None,
    offset: int | None = None,
    limit: int | None = None,
    hidden: bool | None = None,
    get_ancestors: bool | None = None,
    cross_cluster: bool | None = None,
    error_msg: str | None = None,
):
    return connection.get(
        endpoint='/api/searches/results',
        headers={'X-MSTR-ProjectID': project_id},
        params={
            'name': name,
            'type': object_types,
            'pattern': pattern,
            'root': root,
            'offset': offset,
            'limit': limit,
            'getAncestors': get_ancestors,
            'certifiedStatus': certified_status,
            'result.hidden': hidden,
            'isCrossCluster': cross_cluster,
        },
    )


@ErrorHandler(
    err_msg="Error getting quick search result from search object with ID {"
    "search_object_id}"
)
def get_quick_search_result_from_object(
    connection: 'Connection',
    project_id: str,
    search_object_id: str,
    subtypes: (
        None | Union['ObjectSubTypes', list['ObjectSubTypes'], int, list[int]]
    ) = None,
    include_ancestors: bool | None = None,
    include_acl: bool | None = None,
    limit: int | None = None,
    offset: int | None = None,
    error_msg: str | None = None,
):
    return connection.get(
        endpoint=f'/api/searchObjects/{search_object_id}/results',
        headers={'X-MSTR-ProjectID': project_id},
        params={
            'includeAncestors': include_ancestors,
            'includeAcl': include_acl,
            'result.subtypes': subtypes,
            'offset': offset,
            'limit': limit,
        },
    )


@ErrorHandler(err_msg="Error getting specified shortcuts.")
def get_shortcuts(
    connection: 'Connection',
    body: dict,
    shortcut_info_flag: int = 0,
    error_msg: str | None = None,
):
    """Retrieve information about specific published shortcuts
    in specific projects.

    Args:
        connection: Strategy One REST API connection object
        body: A dictionary specifying the projects and shortcuts in the form of
            {[
                {
                    'projectId': 'string',
                    'shortcutIds': [
                    'string'
                    ]
                }
            ]}.
        shortcut_info_flag: flag indicating what information about shortcut
                should be loaded
        error_msg: Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """

    return connection.post(
        endpoint=(
            f'/api/searches/library/shortcuts?shortcutInfoFlag={shortcut_info_flag}'
        ),
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(err_msg="Error getting shortcut with id {id}.")
def get_shortcut(
    connection: 'Connection',
    id: str,
    project_id: str,
    shortcut_info_flag: int = 2,
    error_msg: str | None = None,
):
    """Get information about specific published shortcut in specific project.

    Args:
        connection: Strategy One REST API connection object
        id: id of target shortcut
        project_id: id of project that the shortcut is in
        shortcut_info_flag: flag indicating what information about shortcut
                should be loaded
        error_msg: Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id

    response = connection.post(
        endpoint=(
            f'/api/searches/library/shortcuts?shortcutInfoFlag={shortcut_info_flag}'
        ),
        headers={'X-MSTR-ProjectID': None},
        json=[{'projectId': project_id, 'shortcutIds': [id]}],
    )

    if response.ok:
        response_json = response.json()
        if len(response_json) > 0:
            response_json = response_json[0]
        else:
            raise ValueError('Shortcut not found.')
        response.encoding, response._content = 'utf-8', json.dumps(
            response_json
        ).encode('utf-8')
    return response


@ErrorHandler(err_msg="Error getting search suggestions.")
def get_search_suggestions(
    connection: 'Connection',
    project_id: str | None = None,
    key: str | None = None,
    count: int = -1,
    is_cross_cluster: bool = None,
):
    """Store results of the Search engine to return search suggestions.

    Args:
        connection (object): Strategy One REST API connection object
        project_id (string, optional): project ID
        key (string, optional): value the search pattern is set to, which will
            be applied to the names of suggestions being searched
        count (int, optional): maximum number of items returned for a single
            request. Used to control paging behavior. Default value is `-1` for
            no limit.
        is_cross_cluster (bool, optional): perform search in all unique projects
            across the cluster, this parameter only takes effect for I-Server
            with cluster nodes. Default value is `None`
    """
    is_cross_cluster = str(is_cross_cluster).lower() if is_cross_cluster else None
    return connection.get(
        endpoint=(
            '/api/searches/suggestions'
            f'?key={key}?count={count}?isCrossCluster={is_cross_cluster}'
        ),
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(err_msg="Error getting specified objects.")
def get_objects_from_quick_search(
    connection: 'Connection',
    body: dict,
    include_ancestors: bool = False,
    show_navigation_path: bool = False,
    fields: str | None = None,
    error_msg: str | None = None,
):
    """Search for objects in the metadata.

    Args:
        connection: Strategy One REST API connection object
        body: A dictionary specifying the search criteria
        include_ancestors: Include ancestors in the search results
        show_navigation_path: Show navigation path in the search results
        fields: Comma-separated list of fields to include in the search results
        error_msg: Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.post(
        endpoint='/api/searches/objects',
        json=body,
        params={
            'includeAncestors': include_ancestors,
            'showNavigationPath': show_navigation_path,
            'fields': fields,
        },
    )


@ErrorHandler(err_msg="Error getting search object with ID {id}.")
def get_search_object(
    connection: 'Connection',
    id: str,
    error_msg: str | None = None,
):
    """Get information about a specific search object.

    Args:
        connection: Strategy One REST API connection object
        id: ID of the search object
        error_msg: Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint=f'/api/searchObjects/{id}',
    )


@ErrorHandler(err_msg="Error creating search object.")
def create_search_object(
    connection: 'Connection',
    project_id: str,
    body: dict,
    error_msg: str | None = None,
):
    """Create a search object.

    Args:
        connection: Strategy One REST API connection object
        project_id: ID of the project where the search object will be created
        body: Dictionary containing the search object details
        error_msg: Custom Error Message for Error Handling

    Returns:
        HTTP response object.
    """

    return connection.post(
        endpoint='/api/searchObjects',
        headers={'X-MSTR-ProjectID': project_id},
        json=body,
    )


@ErrorHandler(err_msg="Error updating search object with ID {id}.")
def update_search_object(
    connection: 'Connection',
    id: str,
    body: dict,
    error_msg: str | None = None,
):
    """
    Alter a stored search object.

    Args:
        connection (object): Strategy One connection object returned by
            `connection.Connection()`.
        id (string): Search ID (identifies the search instance to alter)
        body (dict): Dictionary containing the search object details
        error_msg (string, optional): Custom error message for error handling

    Returns:
        HTTP response returned by the Strategy One REST server.
    """

    return connection.patch(
        endpoint=f'/api/searchObjects/{id}',
        json=body,
    )



"""This is the demo script to show how to manage report caches.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.project_objects import ContentCache, Report
from mstrio.connection import get_connection

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name # Insert project ID here

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Get the project ID
PROJECT_ID = conn.project_id

# List all report caches for a given project
# This can be done in two ways
# Listing it through the Report class
Report.list_caches(connection=conn, project_id=PROJECT_ID)

# Listing it through the ContentCache class
# We have to specify the type of the cache we want to list in this case
ContentCache.list_caches(connection=conn, project_id=PROJECT_ID, type='report')

# Define a variable which can be later used in a script
REPORT_CACHE_ID = $cache_id # Insert ID of cache that you want to perform your actions upon

# Initializing the cache
REPORT_CACHE = ContentCache(connection=conn, id=REPORT_CACHE_ID)

# Listing properties of a cache
REPORT_CACHE.list_properties()

# Invalidating specific caches for reports
Report.invalidate_caches(connection=conn, ids=[REPORT_CACHE_ID])

# Deleting all caches for all reports
Report.delete_all_caches(connection=conn, force=True)

# Deleting a specific cache
REPORT_CACHE.delete(force=True)

# Deleting all caches from a specific warehouse table
WAREHOUSE_TABLE = $warehouse_table # Insert warehouse table name here
cache_ids = [cache.id for cache in Report.list_caches(connection=conn, project_id=PROJECT_ID, wh_tables=WAREHOUSE_TABLE)]
ContentCache.delete_caches(connection=conn, cache_ids=cache_ids, force=True)

# Loading and unloading all caches in a project
cache_ids = [cache.id for cache in Report.list_caches(connection=conn, project_id=PROJECT_ID)]
ContentCache.load_caches(connection=conn, cache_ids=cache_ids)
ContentCache.unload_caches(connection=conn, cache_ids=cache_ids)

# Deleting invalid caches from a project
cache_ids = [cache.id for cache in Report.list_caches(connection=conn, project_id=PROJECT_ID, status='invalid')]
ContentCache.delete_caches(connection=conn, cache_ids=cache_ids, force=True)

# Deleting invalid caches from multiple projects
PROJECT_ID_2 = $project_id_2 # Insert project ID here
PROJECTS = [PROJECT_ID, PROJECT_ID_2]
for project in PROJECTS:
    cache_ids = [cache.id for cache in Report.list_caches(connection=conn, project_id=project, status='invalid')]
    ContentCache.delete_caches(connection=conn, cache_ids=cache_ids, force=True)



import logging
from typing import TYPE_CHECKING

from mstrio import config
from mstrio.api import subscriptions as subscriptions_
from mstrio.connection import Connection
from mstrio.utils import helper
from mstrio.utils.enum_helper import get_enum_val
from mstrio.utils.resolvers import (
    get_project_id_from_params_set,
    validate_owner_key_in_filters,
)
from mstrio.utils.version_helper import (
    class_version_handler,
    is_server_min_version,
    method_version_handler,
)

from . import (
    CacheUpdateSubscription,
    EmailSubscription,
    FileSubscription,
    FTPSubscription,
    HistoryListSubscription,
    MobileSubscription,
    Subscription,
)
from .content import Content
from .delivery import Delivery

if TYPE_CHECKING:
    from mstrio.server.project import Project

logger = logging.getLogger(__name__)


@method_version_handler('11.2.0203')
def list_subscriptions(
    connection: Connection,
    project: 'Project | str | None' = None,
    project_id: str | None = None,
    project_name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    last_run: bool = False,
    **filters,
) -> list["Subscription"] | list[dict]:
    """Get all subscriptions per project as list of Subscription objects or
    dictionaries.

    Optionally filter the subscriptions by specifying filters.

    Args:
        connection (Connection): Strategy One connection object
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        to_dictionary (bool): If True returns a list of subscription dicts,
            otherwise (default) returns a list of subscription objects
        limit (int | None): limit the number of elements returned. If `None`
            (default), all objects are returned.
        last_run (bool): If True, adds the last time that the subscription ran.
        **filters: Available filter parameters: ['id', 'multiple_contents',
            'name', 'editable', 'allow_delivery_changes'
            'allow_personalization_changes', 'allow_unsubscribe',
            'date_created', 'date_modified', 'owner', 'delivery']
    """

    proj_id = get_project_id_from_params_set(
        connection,
        project,
        project_id,
        project_name,
    )
    chunk_size = 1000 if is_server_min_version(connection, '11.3.0300') else 1000000

    validate_owner_key_in_filters(filters)

    if (
        not is_server_min_version(connection, '11.4.0600')
        and last_run
        and config.verbose
    ):
        logger.info('`last_run` argument is available from iServer Version 11.4.0600')

    msg = 'Error getting subscription list.'
    objects = helper.fetch_objects_async(
        connection=connection,
        api=subscriptions_.list_subscriptions,
        async_api=subscriptions_.list_subscriptions_async,
        limit=limit,
        chunk_size=chunk_size,
        filters=filters,
        error_msg=msg,
        dict_unpack_value="subscriptions",
        project_id=proj_id,
        last_run=last_run,
    )

    if to_dictionary:
        return objects
    return [
        dispatch_from_dict(
            source=obj,
            connection=connection,
            project_id=proj_id,
        )
        for obj in objects
    ]


DeliveryMode = Delivery.DeliveryMode
subscription_type_from_delivery_mode_dict = {
    DeliveryMode.CACHE: CacheUpdateSubscription,
    DeliveryMode.EMAIL: EmailSubscription,
    DeliveryMode.FILE: FileSubscription,
    DeliveryMode.FTP: FTPSubscription,
    DeliveryMode.HISTORY_LIST: HistoryListSubscription,
    DeliveryMode.MOBILE: MobileSubscription,
}


def get_subscription_type_from_delivery_mode(mode: DeliveryMode):
    """Returns the subscription type of the provided Delivery Mode.

    Args:
        mode: DeliveryMode object of which to get the subscription type"""
    return subscription_type_from_delivery_mode_dict.get(mode, Subscription)


def dispatch_from_dict(
    source: dict, connection: Connection, project_id: str
) -> 'Subscription':
    """Returns the subscription type object from the provided source

    Args:
        source: dictionary of an object to return from the specified
            subscription
        connection: Strategy One connection object returned
            by `connection.Connection()`
        project_id: Project ID

    Returns:
        Subscription: The subscription type object
    """
    delivery_mode = DeliveryMode(source["delivery"]["mode"])
    sub_type = get_subscription_type_from_delivery_mode(delivery_mode)
    return sub_type.from_dict(source, connection, project_id)


@class_version_handler('11.2.0203')
class SubscriptionManager:
    """Manage subscriptions."""

    def __init__(
        self,
        connection: Connection,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
    ):
        """Initialize the SubscriptionManager object.

        Args:
            connection (Connection): Strategy One connection object returned
                by `connection.Connection()`
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
        """
        self.connection = connection
        self.project_id = get_project_id_from_params_set(
            connection,
            project,
            project_id,
            project_name,
        )

    def list_subscriptions(
        self,
        to_dictionary: bool = False,
        limit: int | None = None,
        last_run: bool = False,
        **filters,
    ):
        """Get all subscriptions as list of Subscription objects or
        dictionaries.

        Optionally filter the subscriptions by specifying filters.

        Args:
            to_dictionary: If True returns a list of subscription dicts,
                otherwise returns a list of subscription objects
            limit: limit the number of elements returned. If `None` (default),
                all objects are returned.
            last_run: If True, adds the last time that the subscription ran.
            **filters: Available filter parameters: ['id', 'name', 'editable',
                'allowDeliveryChanges', 'allowPersonalizationChanges',
                'allowUnsubscribe', 'dateCreated', 'dateModified', 'owner',
                'schedules', 'contents', 'recipients', 'delivery']
        """
        return list_subscriptions(
            connection=self.connection,
            project_id=self.project_id,
            to_dictionary=to_dictionary,
            limit=limit,
            last_run=last_run,
            **filters,
        )

    def _normalize_subscriptions(self, subscriptions) -> list[Subscription]:
        if not isinstance(subscriptions, list):
            subscriptions = [subscriptions]
        return [
            (
                sub
                if isinstance(sub, Subscription)
                else Subscription(
                    connection=self.connection, id=sub, project_id=self.project_id
                )
            )
            for sub in subscriptions
        ]

    def _delete_subscription(self, subscription) -> bool:
        response = subscriptions_.remove_subscription(
            self.connection,
            subscription.id,
            self.project_id,
            error_msg=(
                f"Subscription '{subscription.name}' with id "
                f"'{subscription.id}' could not be deleted."
            ),
            exception_type=UserWarning,
        )
        if response.ok and config.verbose:
            logger.info(
                f"Deleted subscription '{subscription.name}' "
                f"with ID '{subscription.id}'."
            )
        return response.ok

    def create_copy(
        self,
        subscription: Subscription | str,
        name: str | None = None,
        project: 'Project | str | None' = None,
        project_id: str | None = None,
        project_name: str | None = None,
        send_now: bool = False,
    ):
        """Create a copy of the subscription on the I-Server.

        Args:
            subscription (Subscription | str): Subscription object or ID of the
                subscription to be copied
            name (str, optional): New name of the object. If None, a default
                name is generated, such as 'Old Name (1)'
            project (Project | str, optional): Project object or ID or name
                specifying the project. May be used instead of `project_id` or
                `project_name`.
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
            send_now (bool): indicates whether to execute the subscription
                immediately.

        Returns:
            New object, the copied subscription. The subscription's name might
            be changed to avoid conflicts with existing objects.
        """
        # Subscription is not a Metadata object and as such cannot be copied
        # with the objects' copy endpoint.
        # For the same reason, duplicate name has to be determined client-side
        if not isinstance(subscription, Subscription):
            subscription = Subscription(
                connection=self.connection, id=subscription, project_id=self.project_id
            )
        name = name or subscription.name
        existing_names: list[str] = [
            s['name']
            for s in list_subscriptions(
                subscription.connection,
                to_dictionary=True,
                project=project,
                project_id=project_id,
                project_name=project_name,
            )
        ]
        new_name = helper.deduplicated_name(name, existing_names)

        proj_id = get_project_id_from_params_set(
            self.connection,
            project,
            project_id or self.project_id,
            project_name,
        )

        return Subscription._Subscription__create(
            connection=subscription.connection,
            name=new_name,
            contents=subscription.contents,
            project_id=proj_id,
            project_name=project_name,
            multiple_contents=subscription.multiple_contents,
            allow_delivery_changes=subscription.allow_delivery_changes,
            allow_personalization_changes=subscription.allow_personalization_changes,
            allow_unsubscribe=subscription.allow_unsubscribe,
            send_now=send_now,
            owner_id=subscription.owner.id,
            schedules=subscription.schedules,
            recipients=subscription.recipients,
            delivery=subscription.delivery,
        )

    def delete(
        self, subscriptions: list[Subscription] | list[str], force=False
    ) -> bool:
        """Deletes all passed subscriptions. Returns True if successfully
        removed all subscriptions.

        Args:
            subscriptions (list[Subscription] | list[str]):
                list of subscriptions to be deleted
            force (bool, optional): if True skips the prompt asking for
                confirmation before deleting subscriptions. False by default.
        """
        subscriptions = self._normalize_subscriptions(subscriptions)
        if not subscriptions:
            if config.verbose:
                logger.info('No subscriptions passed.')
            return False

        if not force:
            logger.info("Found subscriptions:")
            for sub in subscriptions:
                logger.info(f"Subscription '{sub.name}' with ID: '{sub.id}'")
            if input("Are you sure you want to delete all of them? [Y/N]: ") != 'Y':
                return False

        succeeded = sum(
            1
            for subscription in subscriptions
            if self._delete_subscription(subscription)
        )

        return succeeded == len(subscriptions)

    def execute(self, subscriptions: list[Subscription] | list[str]) -> None:
        """Executes all passed subscriptions.

        Args:
            subscriptions (list[Subscription] | list[str]):
                list of subscription objects or subscription ids to be executed
        """
        if not subscriptions:
            if config.verbose:
                logger.info('No subscriptions passed.')
            return

        subscriptions = (
            subscriptions if isinstance(subscriptions, list) else [subscriptions]
        )
        for subscription in subscriptions:
            if not isinstance(subscription, Subscription):
                subscription = Subscription(
                    connection=self.connection,
                    id=subscription,
                    project_id=self.project_id,
                )

            if subscription.delivery.mode in (
                'EMAIL',
                'FILE',
                'HISTORY_LIST',
                'FTP',
            ):
                subscription.execute()
            else:
                msg = (
                    f"Subscription '{subscription.name}' with ID "
                    f"'{subscription.id}' could not be executed. Delivery mode "
                    f"'{subscription.delivery.mode}' is not supported."
                )
                helper.exception_handler(msg, UserWarning)

    @method_version_handler('11.3.0000')
    def available_bursting_attributes(self, content: dict | Content) -> list[dict]:
        """Get a list of available attributes for bursting feature, for a given
        content.

        Args:
            content (dict | Content): content dictionary or Content object
                (from subscription.content)
        """
        c_id = content.id if isinstance(content, Content) else content.get('id')
        c_type = (
            get_enum_val(content.type, Content.Type)
            if isinstance(content, Content)
            else content.get('type')
        )

        response = subscriptions_.bursting_attributes(
            self.connection, self.project_id, c_id, c_type.upper()
        )

        return response.json()['burstingAttributes']

    @method_version_handler('11.3.0000')
    def available_recipients(
        self,
        content_id: str | None = None,
        content_type: str | Content.Type | None = None,
        content: 'Content | None' = None,
        delivery_type: str | DeliveryMode = 'EMAIL',
    ) -> list[dict]:
        """List available recipients for a subscription contents.
        Specify either both `content_id` and `content_type` or just `content`
        object.

        Args:
            content_id (str): ID of the content.
            content_type (str | Content.Type): type of the content.
            content (Content): Content object.
            delivery_type (str | DeliveryMode): The delivery type
                of the subscription.
        """
        if isinstance(content, Content):
            content_id = content.id
            content_type = content.type
        elif not (content_id and content_type):
            helper.exception_handler(
                'Specify either a content ID and type or content object.', ValueError
            )

        body = {
            "contents": [
                {"id": content_id, "type": get_enum_val(content_type, Content.Type)}
            ],
        }

        response = subscriptions_.available_recipients(
            connection=self.connection,
            project_id=self.project_id,
            body=body,
            delivery_type=get_enum_val(delivery_type, DeliveryMode),
        )

        return response.json()['recipients']



import json
import logging
from abc import ABCMeta, abstractmethod
from pprint import pprint
from sys import version_info
from typing import Any

from pandas import DataFrame, Series

from mstrio import config
from mstrio.utils import helper

from .setting_types import DeprecatedSetting, SettingValue, SettingValueFactory
from .settings_helper import convert_settings_to_byte, convert_settings_to_mega_byte
from .settings_io import (
    CSVSettingsIO,
    JSONSettingsIO,
    PickleSettingsIO,
    SettingsSerializerFactory,
)

logger = logging.getLogger(__name__)


class BaseSettings(metaclass=ABCMeta):
    """Base class for fetching and updating Project or Server Settings. The
    settings object can be optionally initialized with `connection` and `id`,
    which will automatically fetch the current settings for the specified
    object. If not specified, settings can be loaded from file using
    `import_from()` method. Settings attributes can be modified manually.
    Lastly, the settings object can be validated and applied to a given object
    (Project, Server) using the `update()` method.

    Attributes:
        settings: settings
    """

    _CONFIG: dict = {}
    # !NOTE define in child classes
    _TYPE: str | None = None
    _READ_ONLY: list[str] = []
    _CONVERSION_MAP: dict[str, str] = {}

    def alter(self, **settings):
        """Alter settings in a command manager like way."""
        for setting, value in settings.items():
            setattr(self, setting, value)

    def fetch(self) -> None:
        """Fetch current settings from I-Server and update this Settings
        object."""
        prepared_settings = self._fetch()
        for key, value in prepared_settings.items():
            if hasattr(self, key):
                setting_obj = getattr(self, key)
                setting_obj.value = value
                super().__setattr__(key, setting_obj)

    @abstractmethod
    def _fetch(self):
        pass

    @abstractmethod
    def update(self) -> None:
        """Update the current settings on I-Server using this Settings
        object."""
        for key, setting in self.__dict__.items():
            if (
                not key.startswith('_')
                and isinstance(setting, SettingValue)
                and setting._modified
            ):
                setting._modified = False

    def _get_config(self):
        """Fetch the settings config to be used in this Settings class."""
        for setting, cfg in self._CONFIG.items():
            setting_type = cfg.get('type')
            if setting_type is None:
                self._READ_ONLY.append(setting)
            cfg.update({'name': setting})

    def compare_with_files(
        self, files: list[str], show_diff_only: bool = False
    ) -> DataFrame:
        """Compare the current project settings to settings in file/files
        Args:
            files (str): List of paths to the settings files. Supported settings
                files are JSON, CSV, Pickle. Ex: "/file.csv"
            show_diff_only(bool, optional): Whether to display all settings or
                only different from first project in list.
        Returns:
            Dataframe with values of current settings and settings from files.
        """
        current = DataFrame.from_dict(
            self.list_properties(show_names=True), orient='index', columns=['value']
        ).reset_index()
        base = 'current value'
        current.columns = ['setting', base]
        for proj in files:
            self.import_from(proj)
            current[proj] = self.to_dataframe()['value']
        if show_diff_only:
            index = Series([True] * len(current['setting']))
            for proj_name in files:
                compare = current[base].eq(current[proj_name])
                index = compare & index
            current = current[~index]
            if current.empty and config.verbose:
                logger.info(
                    'There is no difference between current settings and settings '
                    'from files.'
                )
        return current

    def to_csv(self, name: str, show_description: bool = False) -> None:
        """Export the current project settings to the csv file.

        Args:
            name (str): Name of file
            show_description (bool): if True return, description and value
                for each setting, else, return values only
        """
        CSVSettingsIO.to_file(
            file=name, settings_obj=self, show_description=show_description
        )

    def to_json(self, name: str) -> None:
        """Export the current project settings to the json file
        Args:
            name (str): Name of file
        """
        JSONSettingsIO.to_file(file=name, settings_obj=self)

    def to_pickle(self, name: str) -> None:
        """Export the current project settings to the pickle file
        Args:
            name (str): Name of file
        """
        PickleSettingsIO.to_file(file=name, settings_obj=self)

    def import_from(self, file: str) -> None:
        """Import project settings from a 'csv', 'json' or 'pickle' file.

        Args:
            file (str): Path to the file with supported extension type name.
                Ex: "<path>/file.csv"
        """
        # Extract settings from supported file
        serializer = SettingsSerializerFactory().get_serializer(file)
        settings_dict = serializer.from_file(file, self)

        # Try to validate settings by fetching settings from I-Server
        self._validate_settings(settings_dict)

        # If no Exception was raised, assign the settings to the object
        for setting, value in settings_dict.items():
            setattr(self, setting, value)
        if config.verbose:
            logger.info(f"Settings imported from '{file}'")

    def list_properties(
        self, show_names: bool = True, show_description: bool = False
    ) -> dict:
        """Return settings and their values as dictionary.

        Args:
            show_names (bool): if True, return meaningful setting values,
                else, return exact setting values
            show_description (bool): if True return, description and value
                for each setting, else, return values only
        """
        return {
            key: (
                {
                    'value': (
                        self.__dict__[key]._get_value()
                        if show_names
                        else self.__dict__[key].value
                    ),
                    'description': self.__dict__[key].description,
                }
                if show_description
                else (
                    self.__dict__[key]._get_value()
                    if show_names
                    else self.__dict__[key].value
                )
            )
            for key in sorted(self.__dict__)
            if not key.startswith('_')
        }

    def to_dataframe(self) -> DataFrame:
        """Return a `DataFrame` object containing settings and their values."""

        df = DataFrame.from_dict(
            self.list_properties(), orient='index', columns=['value']
        )
        df.reset_index(inplace=True)
        df.rename({'index': 'setting'}, axis=1, inplace=True)
        return df

    @property
    def info(self) -> None:
        if version_info.major >= 3 and version_info.minor >= 7:
            pprint(self.list_properties(), sort_dicts=False)
        else:
            pprint(self.list_properties())

    @property
    def setting_types(self) -> DataFrame:
        df = DataFrame.from_dict(self._CONFIG, orient='index')
        df = df.drop(
            ['name', 'read_only', 'reboot_rule', 'deprecated'],
            axis='columns',
            errors='ignore',
        )
        return df

    def _validate_settings(
        self, settings: dict | None = None, bad_setting=Warning, bulk_error=True
    ) -> None:
        """Validate setting-value pairs and raise AttributeError or TypeError
        if invalid. If `bad_setting` or `bad_type` is of type Exception, then
        Exception is raised as soon as the first invalid pair is found. If they
        are of type Warning the validation will continue and not raise error.

        Raises:
            ValueError if `bulk_error` True, tries to evaluate all settings.
        """
        settings = settings if settings else self.list_properties(show_names=False)
        bad_settings_keys = []

        for setting, value in settings.items():
            if setting not in self._CONFIG.keys():
                msg = f"Setting '{setting}' is not supported."
                helper.exception_handler(msg, bad_setting)
                bad_settings_keys.append((setting, value))
            else:
                setting_obj = getattr(self, setting)
                if isinstance(setting_obj, DeprecatedSetting):
                    continue
                else:
                    valid = setting_obj._validate_value(value, exception=not bulk_error)
                    if not valid:
                        bad_settings_keys.append((setting, value))
        if bulk_error and bad_settings_keys:
            helper.exception_handler(
                "Invalid settings: {}".format(
                    [item[0] + ': ' + str(item[1]) for item in bad_settings_keys]
                ),
                exception_type=ValueError,
            )

    def _prepare_settings_push(self) -> dict:
        def to_rest_format(settings: dict) -> dict:
            return {
                k: {'value': v} for k, v in settings.items() if k not in self._READ_ONLY
            }

        settings = self.list_properties(show_names=False)
        settings = {
            k: v for k, v in settings.items() if k in set(self._get_modified_settings())
        }
        settings = convert_settings_to_byte(settings, self._CONVERSION_MAP)
        settings = to_rest_format(settings)
        return settings

    def _prepare_settings_fetch(self, settings: dict) -> dict:
        def from_rest_format(settings: dict) -> dict:
            return {k: v['value'] for k, v in settings.items()}

        settings = from_rest_format(settings)
        settings = convert_settings_to_mega_byte(settings, self._CONVERSION_MAP)
        return settings

    def _cast_type_from_obj(self, setting: str, value: Any) -> Any:
        setting_obj = getattr(self, setting, None)
        processed_value = None

        if setting_obj and not isinstance(setting_obj, DeprecatedSetting):
            try:
                processed_value = setting_obj.type(value)
            except ValueError:
                processed_value = value
        else:
            processed_value = value
        return processed_value

    def _configure_settings(self):
        """Sets up the settings object to allow for verification of
        settings."""

        self._get_config()
        for setting, cfg in self._CONFIG.items():
            factory = SettingValueFactory()
            value = factory.get_setting(cfg)
            self.__override_settings_config(value)
            super().__setattr__(setting, value)

    def __override_settings_config(self, value: SettingValue) -> None:
        # placeholder to be overwritten
        return

    def __str__(self):
        """String interpretation of the Setting object."""
        return json.dumps((self.list_properties()), indent=4)

    def __setattr__(self, name, value):
        setting = self.__getattribute__(name)
        if isinstance(setting, (SettingValue, DeprecatedSetting)):
            if setting.value != value or not isinstance(setting.value, type(value)):
                setting._validate_value(value)
                if getattr(setting, 'reboot_rule', None) is not None:
                    msg = f"{setting.name}: {setting.reboot_rule.get('description')}"
                    helper.exception_handler(msg=msg, exception_type=Warning)
                setting.value = value
                if not isinstance(setting, DeprecatedSetting):
                    setting._modified = True
        else:
            super().__setattr__(setting, value)

    def _get_modified_settings(self):
        return [
            key
            for key, setting in self.__dict__.items()
            if not key.startswith('_')
            and isinstance(setting, SettingValue)
            and setting._modified
        ]



from mstrio.api import cubes as cubes_api
from mstrio.connection import Connection
from mstrio.modeling.expression import ExpressionFormat
from mstrio.utils.helper import rename_dict_keys

REST_ATTRIBUTES_MAP = {
    'objectId': 'id',
    'timeBased': 'timeBasedSettings',
}


def create(
    connection: Connection,
    body: dict,
    project_id: str | None = None,
    cube_template_id: str | None = None,
    show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
    show_filter_tokens: bool = False,
    show_advanced_properties: bool = False,
) -> dict:
    """Create cube from provided body.

    Args:
        connection (Connection): Strategy One REST API connection object.
        body (dict): JSON-formatted data used to create cube.
        project_id (str, optional): Project ID.
        cube_template_id (str, optional): If specified, new created cube will
            inherit 'template' of given by ID cube.
        show_expression_as (ExpressionFormat, str, optional): Specify how
            expressions should be presented, default 'ExpressionFormat.TREE'.
            Available values:
            - None.
                (expression is returned in 'text' format)
            - `ExpressionFormat.TREE` or `tree`.
                (expression is returned in `text` and `tree` formats)
            - `ExpressionFormat.TOKENS` or `tokens`.
                (expression is returned in `text` and `tokens` formats)
        show_filter_tokens (bool, optional): Specify whether 'filter'
            is returned in 'tokens' format, along with `text` and `tree`
            formats, default False.
            - If omitted or False, only `text` and `tree` formats are returned.
            - If True, all `text`, 'tree' and `tokens` formats are returned.
        show_advanced_properties (bool, optional): Specify whether to retrieve
            the values of the advanced properties. If omitted or false, nothing
            will be returned for the advanced properties, default False.

    Returns:
        Dict representing created cube object.

    Raises:
        AttributeError: If project ID is not provided and Connection object
            doesn't have project selected.
    """

    data = cubes_api.create_cube(
        connection,
        body,
        project_id=project_id,
        cube_template_id=cube_template_id,
        show_expression_as=show_expression_as,
        show_filter_tokens=show_filter_tokens,
        show_advanced_properties=show_advanced_properties,
    ).json()
    data = {**data, **data['information']}
    return rename_dict_keys(data, REST_ATTRIBUTES_MAP)


def get(
    connection: Connection,
    id: str,
    project_id: str | None = None,
    show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
    show_filter_tokens: bool = False,
    show_advanced_properties: bool = False,
) -> dict:
    """Get cube by specific ID.

    Args:
        connection (Connection): Strategy One REST API connection object.
        id (str): Cube ID.
        project_id (str, optional): Project ID.
        show_expression_as (ExpressionFormat, str, optional): Specify how
            expressions should be presented, default 'ExpressionFormat.TREE'.
            Available values:
            - None.
                (expression is returned in 'text' format)
            - `ExpressionFormat.TREE` or `tree`.
                (expression is returned in `text` and `tree` formats)
            - `ExpressionFormat.TOKENS` or `tokens`.
                (expression is returned in `text` and `tokens` formats)
        show_filter_tokens (bool, optional): Specify whether 'filter'
            is returned in 'tokens' format, along with `text` and `tree`
            formats, default False.
            - If omitted or False, only `text` and `tree` formats are returned.
            - If True, all `text`, 'tree' and `tokens` formats are returned.
        show_advanced_properties (bool, optional): Specify whether to retrieve
            the values of the advanced properties. If omitted or false, nothing
            will be returned for the advanced properties, default False.

    Returns:
        Dict representing fetched cube object.

    Raises:
        AttributeError: If project ID is not provided and Connection object
            doesn't have project selected.
    """

    data = cubes_api.get_cube(
        connection,
        id,
        project_id=project_id,
        show_expression_as=show_expression_as,
        show_filter_tokens=show_filter_tokens,
        show_advanced_properties=show_advanced_properties,
    ).json()
    return rename_dict_keys(data, REST_ATTRIBUTES_MAP)


def update(
    connection: Connection,
    id: str,
    body: dict,
    project_id: str | None = None,
    show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
    show_filter_tokens: bool = False,
    show_advanced_properties: bool = False,
) -> dict:
    """Update cube specified by ID and with provided body.

    Args:
        connection (Connection): Strategy One REST API connection object.
        id (str): Cube ID.
        body (dict): JSON-formatted data used to update cube.
        project_id (str, optional): Project ID.
        show_expression_as (ExpressionFormat, str, optional): Specify how
            expressions should be presented, default 'ExpressionFormat.TREE'.
            Available values:
            - None.
                (expression is returned in 'text' format)
            - `ExpressionFormat.TREE` or `tree`.
                (expression is returned in `text` and `tree` formats)
            - `ExpressionFormat.TOKENS` or `tokens`.
                (expression is returned in `text` and `tokens` formats)
        show_filter_tokens (bool, optional): Specify whether 'filter'
            is returned in 'tokens' format, along with `text` and `tree`
            formats, default False.
            - If omitted or False, only `text` and `tree` formats are returned.
            - If True, all `text`, 'tree' and `tokens` formats are returned.
        show_advanced_properties (bool, optional): Specify whether to retrieve
            the values of the advanced properties. If omitted or false, nothing
            will be returned for the advanced properties, default False.

    Returns:
        Dict representing updated cube object.

    Raises:
        AttributeError: If project ID is not provided and Connection object
            doesn't have project selected.
    """

    body['timeBased'] = body.get('timeBasedSettings', {})

    data = cubes_api.update_cube(
        connection=connection,
        id=id,
        body=body,
        project_id=project_id,
        show_expression_as=show_expression_as,
        show_filter_tokens=show_filter_tokens,
        show_advanced_properties=show_advanced_properties,
    ).json()

    return rename_dict_keys(data, REST_ATTRIBUTES_MAP)


def get_info(connection: Connection, id: str) -> dict:
    """Get information for specific cube. This request updates
    size, status, path, owner ID and server mode of the cube.

    Args:
        connection: Strategy One REST API connection object.
        id (str): ID of the cube.

    Returns:
        Dict representing cube information.
    """

    data = cubes_api.cube_info(connection, id).json()['cubesInfos'][0]
    return rename_dict_keys(data, REST_ATTRIBUTES_MAP)



"""
Create a dataframe from a Cube.
List available cubes, retrieve their attributes and metrics. Operate with given
cube.

1. Connect to the environment using data from workstation
2. List available cubes (with limit of 10)
3. Get cube with given id and check its available attributes and metrics
4. Choose attributes and metrics for filtering
5. Publish cube before applying filters
6. Apply filters to the cube specifying attributes and metrics
7. Create a dataframe from the filtered cube
8. Print first few rows of newly created dataframe
"""

from time import sleep
from mstrio.connection import get_connection
from mstrio.project_objects.datasets import list_all_cubes, OlapCube

# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
connection = get_connection(workstationData, project_name='MicroStrategy Tutorial')

# List available Cubes (limited to 10)
cubes = list_all_cubes(connection, limit=10)
for cube in cubes:
    print(cube)

# Check available attributes and metrics of a Cube
sample_cube_id = '42FF415D4E162846C87D4FAD8B26BF4E'
sample_cube = OlapCube(connection, id=sample_cube_id)
attributes = sample_cube.attributes
metrics = sample_cube.metrics
print(f"Attributes: {attributes}\nMetrics: {metrics}")

# Choose attributes and metrics
call_center_attribute = attributes[0].get('id')
category_attribute = attributes[1].get('id')

profit_metric = metrics[0].get('id')
profit_margin_metric = metrics[1].get('id')

# Cube needs to be published before we can apply filters
sample_cube.publish()
sleep(5)

# Filter which attributes and metrics to use in a dataframe
sample_cube.apply_filters(
    attributes=[call_center_attribute, category_attribute],
    metrics=[profit_metric, profit_margin_metric],
)

# Create a dataframe from a Cube
dataframe = sample_cube.to_dataframe()
print(dataframe.head())



from requests import Response

from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg="Error creating Content Group.")
def create_content_group(
    connection: Connection, body: dict, fields: str | None = None
) -> Response:
    """Create a new content group.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        body (dict): JSON-formatted body of the new content group
        fields (str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    return connection.post(
        endpoint='/api/contentGroups', json=body, params={'fields': fields}
    )


@ErrorHandler(err_msg="Error getting Content Group with ID: {id}.")
def get_content_group(
    connection: Connection, id: str, fields: str | None = None
) -> Response:
    """Get a content group by its ID.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        id (str): ID of the content group to get
        fields (str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    return connection.get(
        endpoint=f'/api/contentGroups/{id}', params={'fields': fields}
    )


@ErrorHandler(err_msg="Error updating Content Group with ID: {id}.")
def update_content_group(
    connection: Connection, id: str, body: dict, fields: str | None = None
) -> Response:
    """Update a content group.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        id (str): ID of the content group to update
        body (dict): JSON-formatted body of the updated content group
        fields (str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    return connection.patch(
        endpoint=f'/api/contentGroups/{id}', json=body, params={'fields': fields}
    )


@ErrorHandler(err_msg="Error getting Content Groups.")
def list_content_groups(connection: Connection, fields: str | None = None) -> Response:
    """Get a list of content groups.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        fields (str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    return connection.get(endpoint='/api/contentGroups', params={'fields': fields})


@ErrorHandler(err_msg="Error getting content of Content Group with ID: {id}.")
def get_content_group_contents(
    connection: Connection, id: str, project_ids: list[str], fields: str | None = None
) -> Response:
    """Get the content of a content group.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        id (str): ID of the content group to get the content of
        project_id (list[str]): ID of the project the content group belongs to
        fields (str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    return connection.get(
        endpoint=f'/api/contentGroups/{id}/contents',
        params={'projectId': project_ids, 'fields': fields},
    )


@ErrorHandler(err_msg="Error updating content of Content Group with ID: {id}.")
def update_content_group_contents(
    connection: Connection, id: str, body: dict, fields: str | None = None
) -> Response:
    """Update the content of a content group.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        id (str): ID of the content group to update the content of
        body (dict): JSON-formatted body of the updated content group
        fields (str, optional): Comma separated top-level field whitelist. This
            allows client to selectively retrieve part of the response model,
            defaults to None

    Returns:
        HTTP response object returned by the Strategy One REST server.
    """
    return connection.patch(
        endpoint=f'/api/contentGroups/{id}/contents',
        json=body,
        params={'fields': fields},
    )



"""
Disable users who have been inactive for a certain period of time

1. Connect to the environment using data from workstation
2a. Get users who have been inactive for 90 days
2b. Get users who have been inactive for 90 days using data
    from 'Platform Analytics'
3. Get users who have been inactive for 85 days
4. Send a notification to users who have been inactive for 85 days
5. Disable users who have been inactive for 90 days
"""

from datetime import datetime, timezone

from mstrio.connection import Connection, get_connection
from mstrio.datasources import DatasourceInstance
from mstrio.distribution_services.email import send_email
from mstrio.server import Project
from mstrio.users_and_groups import User, list_users


def get_inactive_users_from_database(
    connection: Connection, days_inactive: int
) -> list[User]:
    """Get users who have been inactive for a certain period of time.

    Args:
        connection (Connection): Strategy One connection object.
        days_inactive (int): Number of days to consider a user inactive.
    """

    project = Project(connection=conn, name='Platform Analytics')
    dsi = DatasourceInstance(conn, name='Platform Analytics')

    # Action_type_id 103 corresponds to user logout event
    query = """
    SELECT B.mstr_user_guid, MAX(A.local_timestamp) as max_timestamp
    FROM fact_access_transactions_view AS A
    JOIN lu_account AS B
    ON A.account_id = B.account_id AND A.action_type_id = '103'
    GROUP BY B.mstr_user_guid
    """

    res = dsi.execute_query(project_id=project.id, query=query)['results']['data']
    # Create a dictionary mapping user IDs to their last login timestamp
    user_last_login = dict(zip(res['mstr_user_guid'], res['max_timestamp']))

    users = list_users(connection)
    inactive_users = []
    today_date = datetime.now(tz=timezone.utc)
    for user in users:
        last_login = user_last_login.get(user.id)
        if user.enabled and last_login:
            last_login = datetime.fromisoformat(last_login)
            result = today_date - last_login
            if result.days > days_inactive:
                inactive_users.append(user)
    return inactive_users


def get_inactive_users(connection: Connection, days_inactive: int) -> list[User]:
    """Get users who have been inactive for a certain period of time.

    Args:
        connection (Connection): Strategy One connection object.
        days_inactive (int): Number of days to consider a user inactive.
    """
    users = list_users(connection)
    today_date = datetime.now(tz=timezone.utc)
    inactive_users = []
    for user in users:
        if user.enabled and user.last_login:
            result = today_date - user.last_login
            if result.days > days_inactive:
                inactive_users.append(user)
    return inactive_users


def disable_inactive_users(users: list[User]) -> None:
    """Disable users who have been inactive for a certain period of time.

    Args:
        users (list[User]): List of users to disable.
    """
    for user in users:
        user.alter(enabled=False)


def send_notification_to_inactive_users(
    connection: Connection, users: list[User]
) -> None:
    """Send a notification to users who have been inactive for a certain
    period of time.

    Args:
        connection (Connection): Strategy One connection object.
        users (list[User]): List of users to send notifications to.
    """

    # Send personalized email to each user
    for user in users:
        inactive_days = datetime.now(tz=timezone.utc) - user.last_login
        send_email(
            connection=connection,
            users=[user],
            subject='Inactive Account',
            content=f'{user.username} account was inactive for {inactive_days} '
            f'and will be disabled soon. Please log in to prevent this.',
        )
    # Alternatively, send a general email to all users
    # send_email(
    #     connection,
    #     users=users,
    #     subject='Inactive Account',
    #     content='Your account was inactive and will be disabled soon. '
    #     'Please log in to prevent this.',
    # )


# Define project to connect to
PROJECT_NAME = 'MicroStrategy Tutorial'

# Connect to environment without providing user credentials
# Variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, project_name=PROJECT_NAME)

# `User.last_login` property is only available on containerized environments
# with the telemetry service enabled
inactive_users_90_days = get_inactive_users(conn, 90)
# For environments without the telemetry service, you can fetch the data from
# DatasourceInstance 'Platform Analytics'
inactive_users_90_days_db = get_inactive_users_from_database(conn, 90)
inactive_users_80_days = get_inactive_users(conn, 85)


send_notification_to_inactive_users(conn, inactive_users_80_days)
disable_inactive_users(inactive_users_90_days)



"""This is the demo script to show how to manage dashboards. Its basic goal is
to present what can be done with this module and to ease its usage.
"""

from mstrio.connection import get_connection
from mstrio.project_objects.dashboard import (
    Dashboard,
    list_dashboards,
    list_dashboards_across_projects
)

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert name of project here

conn = get_connection(workstationData, PROJECT_NAME)

# Define variables which can be later used in a script
PROJECT_ID = $project_id # Insert project ID here
DASHBOARD_NAME = $dashboard_name  # Insert name of the Dashboard here

# Dashboard management
# List dashboards with different conditions
list_of_all_dashboards = list_dashboards(connection=conn)
print(list_of_all_dashboards)
list_of_all_dashboards_as_dicts = list_dashboards(connection=conn, to_dictionary=True)
print(list_of_all_dashboards_as_dicts)
list_of_all_dashboards_as_dataframes = list_dashboards(connection=conn, to_dataframe=True)
print(list_of_all_dashboards_as_dataframes)
list_of_dashboards_by_project_id = list_dashboards(connection=conn, project_id=PROJECT_ID)
print(list_of_dashboards_by_project_id)
list_of_dashboards_by_name = list_dashboards(connection=conn, name=DASHBOARD_NAME)
print(list_of_dashboards_by_name)

# List of dashboards across projects, with examples of different conditions
list_of_dashboards_across_projects = list_dashboards_across_projects(connection=conn)
print(list_of_dashboards_across_projects)
list_of_dashboards_across_projects_as_dicts = list_dashboards_across_projects(
    connection=conn, to_dictionary=True
)
print(list_of_dashboards_across_projects_as_dicts)
list_of_dashboards_across_projects_as_dataframe = list_dashboards_across_projects(
    connection=conn, to_dataframe=True
)
print(list_of_dashboards_across_projects_as_dataframe)

# Define a variable which can be later used in a script
DASHBOARD_ID = $dashboard_id  # Insert ID of Dashboard on which you want to perform actions

# Get single dashboard by its id
dashboard = Dashboard(connection=conn, id=DASHBOARD_ID)

# List dashboard properties
properties = dashboard.list_properties()
print(properties)

# Delete dashboard without prompt
dashboard.delete(force=True)

# Define variables which can be later used in a script
NEW_DASHBOARD_NAME = $new_dashboard_name  # Insert new name of edited dashboard here
NEW_DASHBOARD_DESCRIPTION = $new_dashboard_description  # Insert new description of edited dashboard here
FOLDER_ID = $folder_id  # Insert folder ID here

# Alter dashboard
dashboard.alter(name=NEW_DASHBOARD_NAME)
dashboard.alter(description=NEW_DASHBOARD_DESCRIPTION)
dashboard.alter(folder_id=FOLDER_ID)

# Define a variable which can be later used in a script
USER_ID = $user_id  # Insert user ID here

# Publish and unpublish dashboard
dashboard.publish()
dashboard.publish(recipients=USER_ID)
dashboard.unpublish()

# Share dashboard with given user
dashboard.share_to(users=USER_ID)

# List dashboard schedules
schedules = dashboard.list_available_schedules()
print(schedules)
schedules_as_dicts = dashboard.list_available_schedules(to_dictionary=True)
print(schedules_as_dicts)

# List connected cubes
cubes = dashboard.get_connected_cubes()
print(cubes)

# Get list of dashboard cache, with examples of different conditions
list_of_all_dashboard_cache = Dashboard.list_caches(connection=conn)
print(list_of_all_dashboard_cache)
list_of_limited_dashboard_cache = Dashboard.list_caches(connection=conn, limit=5)
print(list_of_limited_dashboard_cache)

# Define a variable which can be later used in a script
CACHE_ID = $cache_id  # Insert ID of cache on which you want to perform actions

list_of_dashboard_cache_by_id = Dashboard.list_caches(connection=conn, id=CACHE_ID)
print(list_of_dashboard_cache_by_id)
list_of_all_dashboard_caches_as_dicts = Dashboard.list_caches(connection=conn, to_dictionary=True)
print(list_of_all_dashboard_caches_as_dicts)

# Helper function to list all cache status
def show_caches_status():
    caches = Dashboard.list_caches(conn)
    for cache in caches:
        print(f'Cache ID: {cache.id}, {cache.status}')


# List cache properties
cache = Dashboard.list_caches(connection=conn)[0]
properties = cache.list_properties()
print(properties)

# Define a variable which can be later used in a script
OTHER_CACHE_ID = $other_cache_id  # Insert ID of cache on which you want to perform actions

# Unload multiple dashboard caches
Dashboard.unload_caches(connection=conn, cache_ids=[CACHE_ID, OTHER_CACHE_ID])
show_caches_status()

# Define variables which can be later used in a script
CACHE_STATUS = $cache_status # Insert cache status here
USER_NAME = $user_name # Insert user name here

# Unload all dashboard caches with filter examples
Dashboard.unload_all_caches(connection=conn)
show_caches_status()
Dashboard.unload_all_caches(connection=conn, owner=USER_NAME)
show_caches_status()
Dashboard.unload_all_caches(connection=conn, status=CACHE_STATUS)
show_caches_status()

# Load multiple dashboard caches
Dashboard.load_caches(connection=conn, cache_ids=[CACHE_ID, OTHER_CACHE_ID])
show_caches_status()

# Load all dashboard caches with filter examples
Dashboard.load_all_caches(connection=conn)
show_caches_status()
Dashboard.load_all_caches(connection=conn, owner=USER_NAME)
show_caches_status()
Dashboard.load_all_caches(connection=conn, status=CACHE_STATUS)
show_caches_status()

# Invalidating specific caches for dashboards
Dashboard.invalidate_caches(connection=conn, ids=[CACHE_ID])

# Delete multiple dashboard caches
Dashboard.delete_caches(connection=conn, cache_ids=[CACHE_ID, OTHER_CACHE_ID], force=True)

# Delete all dashboard caches with filter examples
Dashboard.delete_all_caches(connection=conn, force=True)
Dashboard.delete_all_caches(connection=conn, owner=USER_NAME, force=True)
Dashboard.delete_all_caches(connection=conn, status=CACHE_STATUS, force=True)



# lst is a vector input
# n is a vector input
# Metric expression example: ScriptFunctionAggregation<_InputNames="lst, n", _ScriptName=”nth_minimal_value”>( Cost, 2){~+}

def nth_minimal_value(lst, n):
    # Sort the list in ascending order
    sorted_lst = sorted(lst)

    # Check if n is within the range of the list length
    if n <= 0 or n > len(sorted_lst):
        raise ValueError("n is out of the valid range")

    # Return the nth minimal value (1-based index)
    return sorted_lst[n-1]

# must-have method which returns the values served as the metric evaluation results.
def get_results():
    return nth_minimal_value($lst, $n)

print(get_results())



import logging
from enum import auto
from typing import TYPE_CHECKING

from mstrio import config
from mstrio.api import administration
from mstrio.connection import Connection
from mstrio.server.project import Project
from mstrio.users_and_groups.user_group import UserGroup
from mstrio.utils.entity import DeleteMixin, EntityBase
from mstrio.utils.enum_helper import AutoName, get_enum_val
from mstrio.utils.helper import fetch_objects, find_object_with_name
from mstrio.utils.version_helper import class_version_handler, method_version_handler

if TYPE_CHECKING:
    from mstrio.users_and_groups.user import User

logger = logging.getLogger(__name__)


class FenceType(AutoName):
    USER_FENCE = auto()
    WORKLOAD_FENCE = auto()


@method_version_handler('11.3.0800')
def list_fences(
    connection: Connection,
    to_dictionary: bool = False,
    limit: int | None = None,
    **filters,
) -> list['Fence'] | list[dict]:
    """Get list of fences.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        to_dictionary (bool, optional): If True returns dicts, by default
            (False) returns objects.
        limit (int, optional): limit the number of elements returned. If `None`
            (default), all objects are returned.
        **filters: Available filter parameters: ['name', 'id', 'type', 'rank']

    Returns:
        A list of fences objects or dictionaries representing them.
    """

    fences = fetch_objects(
        connection=connection,
        api=administration.list_fences,
        to_dictionary=to_dictionary,
        filters=filters,
        limit=limit,
        dict_unpack_value='fences',
    )

    if to_dictionary:
        return fences
    return [Fence.from_dict(fence, connection) for fence in fences]


@class_version_handler('11.3.0800')
class Fence(EntityBase, DeleteMixin):
    """Python representation of a Strategy One Fence object.

    Attributes:
        id (str): ID of the fence.
        name (str): Name of the fence.
        rank (int): Rank or precedence of the fence.
        type (FenceType): Type of fence.
        nodes (list[str]): List of nodes across which the fence is applied.
        users (list[User]): List of users, to which the fence is applied.
        user_groups (list[UserGroup]): List of user groups, to which the fence
            is applied.
        projects (list[Project]): List of projects, to which the fence
            is applied.
    """

    _REST_ATTR_MAP = {
        'usergroups': 'user_groups',
    }

    _API_GETTERS = {
        (
            'id',
            'name',
            'rank',
            'type',
            'nodes',
            'users',
            'user_groups',
            'projects',
        ): administration.get_fence
    }

    _API_DELETE = staticmethod(administration.delete_fence)

    @staticmethod
    def _parse_users(
        source: list[dict], connection: 'Connection', to_snake_case: bool = True
    ):
        from mstrio.users_and_groups.user import User

        return [
            User.from_dict(content, connection, to_snake_case) for content in source
        ]

    _FROM_DICT_MAP = {
        'type': FenceType,
        'users': _parse_users,
        'user_groups': lambda source, connection: [
            UserGroup.from_dict(content, connection) for content in source
        ],
        'projects': lambda source, connection: [
            Project.from_dict(content, connection) for content in source
        ],
    }

    def __init__(
        self,
        connection: Connection,
        id: str | None = None,
        name: str | None = None,
    ) -> None:
        """Initializes a new instance of a Fence class

        Args:
            connection (Connection): Strategy One connection object returned
                by `connection.Connection()`
            id (str, Optional): ID of the fence.
            name (str, Optional): Name of the fence.

        Note:
            Parameter `name` is not used when fetching. If only `name` parameter
            is provided, `id` will be found automatically if such object exists.

        Raises:
            ValueError: if both `id` and `name` are not provided
                or if Fence with the given `name` doesn't exist.
        """

        if not id:
            if name:
                fence = find_object_with_name(
                    connection=connection,
                    cls=self.__class__,
                    name=name,
                    listing_function=list_fences,
                )
                id = fence['id']
            else:
                raise ValueError("Please provide either 'id' or 'name' argument.")

        super().__init__(connection=connection, object_id=id, name=name)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.name = kwargs.get('name')
        self.rank = kwargs.get('rank')
        self.nodes = kwargs.get('nodes')
        self.users = kwargs.get('users')
        self.user_groups = kwargs.get('user_groups')
        self.projects = kwargs.get('projects')
        self._type = kwargs.get('type')

    @staticmethod
    def _prepare_fence_body(
        rank: int | None,
        name: str | None,
        type: FenceType | None,
        nodes: str | list[str] | None,
        users: 'User | list[User] | None',
        user_groups: UserGroup | list[UserGroup] | None,
        projects: Project | list[Project] | None,
    ) -> dict:
        from mstrio.users_and_groups.user import User

        def to_list_or_none(item, cls) -> list | None:
            if isinstance(item, cls):
                return [item]
            return item if isinstance(item, list) else None

        nodes = to_list_or_none(nodes, str)
        users = to_list_or_none(users, User)
        user_groups = to_list_or_none(user_groups, UserGroup)
        projects = to_list_or_none(projects, Project)

        body = {
            'rank': rank,
            'name': name,
            'type': get_enum_val(type, FenceType) if type else None,
            'nodes': nodes,
            'users': (
                [{'id': user.id, 'name': user.name} for user in users]
                if users is not None
                else None
            ),
            'usergroups': (
                [
                    {'id': user_group.id, 'name': user_group.name}
                    for user_group in user_groups
                ]
                if user_groups is not None
                else None
            ),
            'projects': (
                [{'id': project.id, 'name': project.name} for project in projects]
                if projects is not None
                else None
            ),
        }
        return body

    @classmethod
    def create(
        cls,
        connection: Connection,
        rank: int,
        name: str,
        type: FenceType,
        nodes: str | list[str],
        users: 'User | list[User] | None' = None,
        user_groups: UserGroup | list[UserGroup] | None = None,
        projects: Project | list[Project] | None = None,
    ) -> 'Fence':
        """Create a new Fence.

        Args:
            connection (Connection): Strategy One connection object returned
                by `connection.Connection()`
            rank (int): Rank or precedence of the fence.
            name (str): Name of the fence.
            type (FenceType): Type of fence.
            nodes (str | list[str]): List of nodes across which the fence
                is applied.
            users (User | list[User], optional):  List of users, to which the
                fence is applied.
            user_groups (UserGroup | list[UserGroup], optional):
                List of user groups, to which the fence is applied.
            projects (Project | list[Project], optional): List of projects,
                to which the fence is applied.

        Returns:
            Fence object
        """

        if type == FenceType.USER_FENCE and not any([users, user_groups]):
            raise ValueError(
                "At least one of the following parameters must be provided: "
                "`users` or `user_groups`."
            )

        body = cls._prepare_fence_body(
            rank, name, type, nodes, users, user_groups, projects
        )
        fence = cls.from_dict(
            source=administration.create_fence(connection=connection, body=body).json(),
            connection=connection,
        )
        fence.fetch()
        if config.verbose:
            logger.info(
                f"Successfully created Fence named: '{name}' with ID: '{fence.id}'"
            )
        return fence

    def alter(
        self,
        rank: int | None = None,
        nodes: str | list[str] | None = None,
        users: 'User | list[User] | None' = None,
        user_groups: UserGroup | list[UserGroup] | None = None,
        projects: Project | list[Project] | None = None,
    ) -> None:
        """Alter the Fence object.

        Args:
            rank (int, optional): Rank or precedence of the fence.
            nodes (str | list[str], optional): List of nodes across which the
                fence is applied.
            users (User | list[User], optional):  List of users, to which
                the fence is applied.
            user_groups (UserGroup | list[UserGroup], optional):
                List of user groups, to which the fence is applied.
            projects (Project | list[Project], optional): List of projects,
                to which the fence is applied.
        """
        body = self._prepare_fence_body(
            rank, None, None, nodes, users, user_groups, projects
        )
        operation_list = [
            {'op': 'replace', 'path': path, 'value': value}
            for path, value in [
                ('/rank', body.get('rank')),
                ('/nodes', body.get('nodes')),
                ('/users', body.get('users')),
                ('/usergroups', body.get('usergroups')),
                ('/projects', body.get('projects')),
            ]
            if value is not None
        ]

        response = administration.update_fence(
            connection=self.connection,
            id=self.id,
            body={'operationList': operation_list},
        ).json()
        if config.verbose:
            logger.info(f"Successfully altered Fence ID: {self.id}")
        self._set_object_attributes(**response)

    @property
    def type(self) -> FenceType:
        return self._type



import logging
from typing import TYPE_CHECKING

from pandas import DataFrame

from mstrio import config
from mstrio.access_and_security.privilege_mode import PrivilegeMode
from mstrio.access_and_security.security_role import SecurityRole
from mstrio.api import usergroups
from mstrio.connection import Connection
from mstrio.types import ObjectSubTypes
from mstrio.utils.acl import TrusteeACLMixin
from mstrio.utils.entity import DeleteMixin, Entity, ObjectTypes
from mstrio.utils.helper import (
    exception_handler,
    fetch_objects_async,
    filter_list_of_dicts,
    filter_params_for_func,
)
from mstrio.utils.resolvers import validate_owner_key_in_filters
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.response_processors import usergroups as usergroups_processors
from mstrio.utils.version_helper import method_version_handler

if TYPE_CHECKING:
    from mstrio.access_and_security.privilege import Privilege
    from mstrio.modeling.security_filter import SecurityFilter
    from mstrio.server import Project
    from mstrio.users_and_groups.user import User

logger = logging.getLogger(__name__)


def list_user_groups(
    connection: Connection,
    name_begins: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    **filters,
) -> list["UserGroup"]:
    """Get list of User Group objects or User Group dicts. Optionally filter
    the User Groups by specifying 'name_begins' or other filters.

    Optionally use `to_dictionary` or `to_dataframe` to choose output format.
    If `to_dictionary` is True, `to_dataframe` is omitted.

    Wildcards available for name_begins:
        ? - any character
        * - 0 or more of any characters
        e.g. name_begins = ?onny will return Sonny and Tonny

    Args:
        connection: Strategy One connection object returned by
            `connection.Connection()`
        name_begins: Beginning of a User Groups name which we want to list
        to_dictionary: If True returns dict, by default (False) returns
            User Group objects
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['name', 'id', 'type',
            'abbreviation', 'description', 'subtype', 'date_created',
            'date_modified', 'version', 'acg', 'owner', 'ext_type']


    Examples:
        >>> list_user_groups(connection, name_begins='Group',
        >>>                  description='New group')
    """
    return UserGroup._get_user_groups(
        connection=connection,
        name_begins=name_begins,
        to_dictionary=to_dictionary,
        limit=limit,
        **filters,
    )


class UserGroup(Entity, DeleteMixin, TrusteeACLMixin):
    """Object representation of Strategy One User Group object.

    Attributes:
        connection: A Strategy One connection object
        memberships: User Groups that the User Group is a member of
        members: users that are members of User Group
        security_roles: security roles that the User Group is a member of
        privileges: user privileges per project
        id: User ID
        name: User name
        type: Object type
        subtype: Object subtype
        ext_type: Object extended type
        abbreviation: Object abbreviation
        description: Object description
        date_created: Creation time, DateTime object
        date_modified: Last modification time, DateTime object
        version: Object version ID
        owner: Owner name and ID
        ancestors: List of ancestor folders
        settings: Settings of User Group
        acg: Access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: Object access control list
    """

    @staticmethod
    def _parse_members(
        source: list[dict], connection: 'Connection', to_snake_case: bool = True
    ):
        """Parse members from response."""
        from mstrio.users_and_groups.user import User

        return [
            (
                User.from_dict(member, connection, to_snake_case)
                if member['subtype'] == ObjectSubTypes.USER.value
                else UserGroup.from_dict(member, connection, to_snake_case)
            )
            for member in source
        ]

    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        "members": _parse_members,
    }

    _SUPPORTED_PATCH_OPERATIONS = {
        "add": "add",
        "remove": "remove",
        "change": "replace",
    }
    _OBJECT_TYPE = ObjectTypes.USERGROUP
    _API_GETTERS = {
        (
            'id',
            'name',
            'type',
            'subtype',
            'ext_type',
            'abbreviation',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'ancestors',
            'acg',
            'acl',
            'ldapdn',
        ): usergroups.get_user_group_info,
        'comments': objects_processors.get_info,
        'memberships': usergroups.get_memberships,
        'members': usergroups.get_members,
        'security_roles': usergroups.get_security_roles,
        'privileges': usergroups.get_privileges,
        'settings': usergroups_processors.get_settings,
    }
    _API_PATCH: dict = {
        (
            'abbreviation',
            'comments',
            'owner',
        ): (objects_processors.update, 'partial_put'),
        (
            'name',
            'description',
            'memberships',
            'security_roles',
            'members',
            'privileges',
            'ldapdn',
        ): (usergroups_processors.update_user_group_info, 'patch'),
    }

    def __init__(
        self,
        connection: Connection,
        name: str | None = None,
        id: str | None = None,
    ) -> None:
        """Initialize UserGroup object by passing `name` or `id`. When `id` is
        provided (not `None`), `name` is omitted.

        Args:
            connection: Strategy One connection object returned
                by `connection.Connection()`
            name: name of User Group
            id: ID of User Group
        """
        if id is None and name is None:
            exception_handler(
                "Please specify either 'name' or 'id' parameter in the constructor."
            )

        if id is None:
            user_groups = UserGroup._get_user_group_ids(
                connection=connection, name_begins=name, name=name
            )
            if user_groups:
                id = user_groups[0]
            else:
                exception_handler(
                    f"There is no User Group with the given name: '{name}'",
                    exception_type=ValueError,
                )
        super().__init__(connection=connection, object_id=id, name=name)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.ldapdn = kwargs.get("ldapdn")

        self._memberships = kwargs.get('memberships')
        self._members = kwargs.get('members')
        self._security_roles = kwargs.get('security_roles')
        self._privileges = kwargs.get('privileges')
        self._settings = kwargs.get('settings')
        self._security_filters = kwargs.get('security_filters')

    @classmethod
    def create(
        cls,
        connection: Connection,
        name: str,
        description: str | None = None,
        memberships: list[str] | None = None,
        members: list[str] | None = None,
        ldapdn: str | None = None,
    ):
        """Create a new User Group on the I-Server. Returns `UserGroup` object.

        Args:
            connection: Strategy One connection object returned by
                `connection.Connection()`
            name: Name of a newly created User Group
            description: Description of a newly created User Group
            memberships: Specify User Groups which newly created User Group
                will be member
            members: Specify Users which will be members of newly created User
                Group
            ldapdn: User group's LDAP distinguished name
        """
        memberships = memberships or []
        members = members or []
        body = {
            "name": name,
            "description": description,
            "memberships": memberships,
            "members": members,
            "ldapdn": ldapdn,
        }
        response = usergroups.create_user_group(connection, body)
        if response.ok:
            response = response.json()
            if config.verbose:
                logger.info(
                    f"Successfully created user group '{name}' with ID: "
                    f"'{response.get('id')}'"
                )
            return cls.from_dict(source=response, connection=connection)

    @classmethod
    def _get_user_groups(
        cls,
        connection: Connection,
        name_begins: str | None = None,
        to_dictionary: bool = False,
        limit: int | None = None,
        **filters,
    ) -> list["UserGroup"]:
        validate_owner_key_in_filters(filters)

        msg = "Error getting information for a set of User Groups."
        objects = fetch_objects_async(
            connection,
            usergroups.get_info_all_user_groups,
            usergroups.get_info_all_user_groups_async,
            limit=limit,
            chunk_size=1000,
            error_msg=msg,
            name_begins=name_begins,
            filters=filters,
        )
        if to_dictionary:
            return objects
        else:
            return [cls.from_dict(source=obj, connection=connection) for obj in objects]

    @classmethod
    def _get_user_group_ids(
        cls,
        connection: Connection,
        name_begins: str | None = None,
        limit: int | None = None,
        **filters,
    ) -> list[str]:
        group_dicts = UserGroup._get_user_groups(
            connection=connection,
            name_begins=name_begins,
            to_dictionary=True,
            limit=limit,
            **dict(filters),
        )
        return [group['id'] for group in group_dicts]

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        ldapdn: str = None,
        comments: str | None = None,
        owner: 'str | User | None' = None,
    ):
        """Alter User Group name or/and description.

        Args:
            name: New name of the User Group
            description: New description of the User Group
            comments: Long description of the User Group
            owner: owner of the User Group
        """
        from mstrio.users_and_groups.user import User

        if isinstance(owner, User):
            owner = owner.id

        properties = filter_params_for_func(self.alter, locals(), exclude=['self'])
        self._alter_properties(**properties)

    def add_users(self, users: "str | User | list[str | User]") -> None:
        """Add members to the User Group.

        Args:
            users: List of User objects or ids
        """
        succeeded, failed = self._update_nested_properties(users, "members", "add")
        if succeeded and config.verbose:
            logger.info(f"Added {succeeded} user(s) to group {self.name}")
        if failed and config.verbose:
            logger.info(f"User(s) {failed} is/are already a member of '{self.name}'")
        self.fetch('members')

    def remove_users(self, users: "str | User | list[str | User]") -> None:
        """Remove members from User Group.

        Args:
            users: List of User objects or ids
        """
        succeeded, failed = self._update_nested_properties(users, 'members', "remove")
        if succeeded and config.verbose:
            logger.info(f"Removed user(s) '{succeeded}' from group {self.name}")
        if failed and config.verbose:
            logger.warning(f"User(s) {failed} is/are not members of '{self.name}'")
        self.fetch('members')

    def remove_all_users(self) -> None:
        """Remove all members from user group."""
        to_be_removed = [member.id for member in self.members]
        self.remove_users(to_be_removed)

    def list_members(self, **filters) -> list["User | UserGroup"]:
        """List user group members.

        Optionally filter the results by passing filter keyword arguments.

        Args:
            **filters: Available filter parameters: 'name', 'id', 'type',
                'abbreviation', subtype', 'date_created', 'date_modified',
                'version', 'acg', 'owner', source', ext_type', 'username',
                full_name', enabled'
        """
        self.fetch('members')

        if not filters:
            return self.members

        validate_owner_key_in_filters(filters)

        filtered_dicts = filter_list_of_dicts(
            [member.to_dict() for member in self.members], **filters
        )
        filtered_ids = [member['id'] for member in filtered_dicts]
        return [member for member in self.members if member.id in filtered_ids]

    def add_to_user_groups(
        self, groups: "str | UserGroup | list[str | UserGroup]"
    ) -> None:
        """Add User Group to passed groups.

        Args:
            groups: List of User Group objects or ids
        """
        succeeded, failed = self._update_nested_properties(groups, "memberships", "add")
        if succeeded and config.verbose:
            logger.info(f"Added group '{self.name}' to group(s): {succeeded}")
        if failed and config.verbose:
            logger.warning(
                f"Group '{self.name}' is already a member of {failed} group(s)"
            )

    def remove_from_user_groups(
        self, groups: "str | UserGroup | list[str | UserGroup]"
    ) -> None:
        """Remove User Group from passed groups

        Args:
            groups: List of User Group objects or ids
        """
        succeeded, failed = self._update_nested_properties(
            groups, 'memberships', "remove"
        )
        if succeeded and config.verbose:
            logger.info(f"Removed group '{self.name}' from group(s): {succeeded}")
        if failed and config.verbose:
            logger.warning(f"Group '{self.name}' is not a member of {failed} group(s)")

    def grant_privilege(
        self, privilege: "str | list[str] | Privilege | list[Privilege]"
    ) -> None:
        """Grant privileges directly to the User Group.

        Args:
            privilege: List of privilege objects, ids or names
        """
        from mstrio.access_and_security.privilege import Privilege

        privileges = [
            priv['id']
            for priv in Privilege._validate_privileges(self.connection, privilege)
        ]
        existing_ids = [
            privilege['privilege']['id']
            for privilege in self.list_privileges(mode='GRANTED')
        ]
        succeeded, failed = self._update_nested_properties(
            privileges, "privileges", "add", existing_ids
        )

        if succeeded:
            self.fetch('privileges')  # fetch the object privileges
            if config.verbose:
                logger.info(f"Granted privilege(s) {succeeded} to '{self.name}'")
        if failed and config.verbose:
            logger.warning(
                f"User Group '{self.name}' already has privilege(s) {failed}"
            )

    def revoke_privilege(
        self, privilege: "str | list[str] | Privilege | list[Privilege]"
    ) -> None:
        """Revoke directly granted User Group privileges.

        Args:
            privilege: List of privilege objects, ids or names
        """
        from mstrio.access_and_security.privilege import Privilege

        privileges = {
            priv['id']
            for priv in Privilege._validate_privileges(self.connection, privilege)
        }
        existing_ids = [
            privilege['privilege']['id']
            for privilege in self.list_privileges(mode='ALL')
        ]
        directly_granted = {
            privilege['privilege']['id']
            for privilege in self.list_privileges(mode='GRANTED')
        }
        to_revoke = list(privileges.intersection(directly_granted))
        not_directly_granted = list(
            (set(existing_ids) - directly_granted).intersection(privileges)
        )

        if not_directly_granted:
            msg = (
                f"Privileges {sorted(not_directly_granted)} are inherited and will be "
                "omitted. Only directly granted privileges can be revoked by this "
                "method."
            )
            exception_handler(msg, exception_type=Warning)

        succeeded, failed = self._update_nested_properties(
            to_revoke, "privileges", "remove", existing_ids
        )
        if succeeded:
            self.fetch('privileges')  # fetch the object privileges
            if config.verbose:
                logger.info(f"Revoked privilege(s) {succeeded} from '{self.name}'")
        if failed and config.verbose:
            logger.warning(
                f"User group '{self.name}' does not have privilege(s) {failed}"
            )

    def revoke_all_privileges(self, force: bool = False) -> None:
        """Revoke directly granted group privileges.

        Args:
            force: If True, no additional prompt will be shown before revoking
                all privileges from User Group
        """
        user_input = 'N'
        if not force:
            user_input = input(
                "Are you sure you want to revoke all privileges from user group "
                f"'{self.name}'? [Y/N]: "
            )
        if force or user_input == 'Y':
            to_revoke = [
                privilege['privilege']['id']
                for privilege in self.list_privileges(mode='GRANTED')
            ]
            if to_revoke:
                self.revoke_privilege(privilege=to_revoke)
            else:
                logger.info(
                    f"User Group '{self.name}' does not have any directly granted "
                    f"privileges"
                )

    def list_privileges(
        self, mode: PrivilegeMode | str = PrivilegeMode.ALL, to_dataframe: bool = False
    ) -> list:
        """List privileges for user group.

        Args:
            mode: Specifies which privileges to list.
                See: `privilege.PrivilegeMode` enum.
            to_dataframe: If True, return a DataFrame object containing
                privileges
        """
        self.fetch('privileges')

        def to_df(priv_list):
            priv_dict = {}
            for priv in priv_list:
                priv_dict[priv['privilege']['id']] = priv['privilege']['name']
            df = DataFrame.from_dict(priv_dict, orient='index', columns=['Name'])
            df.index.name = 'ID'
            return df

        if not isinstance(mode, PrivilegeMode):
            try:
                mode = PrivilegeMode(mode)
            except ValueError:
                msg = (
                    "Wrong privilege mode has been passed, allowed modes are "
                    "['ALL'/'INHERITED'/'GRANTED']. See: `privilege.PrivilegeMode` "
                    "enum."
                )
                exception_handler(msg, ValueError)

        privileges = []
        if mode == PrivilegeMode.ALL:
            privileges = self.privileges
        elif mode == PrivilegeMode.INHERITED:
            for privilege in self.privileges:
                is_inherited = False
                for source in privilege['sources']:
                    is_inherited = not source['direct'] or is_inherited
                if is_inherited:
                    privileges.append(privilege)
        else:  # GRANTED
            for privilege in self.privileges:
                is_granted = False
                for source in privilege['sources']:
                    is_granted = source['direct'] or is_granted
                if is_granted:
                    privileges.append(privilege)

        return to_df(privileges) if to_dataframe else privileges

    def assign_security_role(
        self,
        security_role: 'SecurityRole | str',
        project: 'Project | str',
    ) -> None:  # NOSONAR
        """Assigns a Security Role to the User Group for given project.

        Args:
            security_role: Security Role ID or object
            project: Project name or object
        """

        security_role = (
            security_role
            if isinstance(security_role, SecurityRole)
            else SecurityRole(self.connection, id=str(security_role))
        )

        security_role.grant_to([self.id], project)
        if config.verbose:
            logger.info(
                f"Assigned Security Role '{security_role.name}' to group: '{self.name}'"
            )
        self.fetch('security_roles')

    def revoke_security_role(
        self, security_role: SecurityRole | str, project: 'Project | str'
    ) -> None:  # NOSONAR
        """Removes a Security Role from the User Group for given project.

        Args:
            security_role: Security Role ID or object
            project: Project name or object
        """

        security_role = (
            security_role
            if isinstance(security_role, SecurityRole)
            else SecurityRole(self.connection, id=str(security_role))
        )

        security_role.revoke_from([self.id], project)
        if config.verbose:
            logger.info(
                f"Revoked Security Role '{security_role.name}' from group: "
                f"'{self.name}'"
            )
        self.fetch('security_roles')

    @method_version_handler('11.3.0200')
    def list_security_filters(
        self, projects: str | list[str] | None = None, to_dictionary: bool = False
    ) -> dict:
        """Get the list of security filters for user group. They can be filtered
        by the projects' ids.

        Args:
            projects (str or list of str, optional): collection of projects' ids
                which is used for filtering data
            to_dictionary: If True returns security filters as dicts, by default
                (False) returns them as objects.

        Returns:
            Dictionary with project names as keys and list with security
            filters as values. In case of no security filter for the given user
            group in the particular project, then this project is not placed in
            the dictionary.
        """
        from mstrio.modeling.security_filter import SecurityFilter

        objects_ = usergroups.get_security_filters(
            self.connection, self.id, projects
        ).json()
        projects_ = objects_.get("projects")

        objects_ = {
            project_.get("name"): project_.get("securityFilters")
            for project_ in projects_
            if project_.get("securityFilters")
        }

        self._security_filters = {
            name: [
                SecurityFilter.from_dict(sec_filter, self.connection)
                for sec_filter in sec_filters
            ]
            for (name, sec_filters) in objects_.items()
        }
        if to_dictionary:
            return objects_
        return self._security_filters

    def apply_security_filter(self, security_filter: "SecurityFilter | str"):
        """Apply a security filter to the user group.

        Args:
            security_filter (string or object): identifier of security filter or
                `SecurityFilter` object which will be applied to the user group.
        Returns:
            True when applying was successful. False otherwise.
        """
        if isinstance(security_filter, str):
            from mstrio.modeling.security_filter import SecurityFilter

            security_filter = SecurityFilter.from_dict(
                {"id": security_filter}, self.connection
            )
        return security_filter.apply(self.id)

    def revoke_security_filter(self, security_filter: "SecurityFilter | str"):
        """Revoke a security filter from the user group.

        Args:
            security_filter (string or object): identifier of security filter or
                `SecurityFilter` object which will be revoked from the user
                group.

        Returns:
            True when revoking was successful. False otherwise.
        """
        if isinstance(security_filter, str):
            from mstrio.modeling.security_filter import SecurityFilter

            security_filter = SecurityFilter.from_dict(
                {"id": security_filter}, self.connection
            )
        return security_filter.revoke(self.id)

    def get_settings(self) -> dict:
        """Get the User Group settings from the I-Server."""
        res = usergroups.get_settings(self.connection, self.id, include_access=True)
        return res.json()

    @property
    def memberships(self):
        return self._memberships

    @property
    def members(self):
        return self._members

    @property
    def security_roles(self):
        return self._security_roles

    @property
    def privileges(self):
        self.fetch('privileges')
        return self._privileges

    @property
    def settings(self) -> dict | list[dict]:
        """Get the User Group settings from the I-Server.

        Returns:
            List of dictionaries with settings on the server version 11.4.1200
            and above.
            Otherwise, returns a dictionary with settings for `jobMemGoverning`.
        """
        return self._settings

    @property
    def security_filters(self):
        if not self._security_filters:
            self.list_security_filters()
        return self._security_filters



from enum import Enum, IntFlag

from mstrio.utils.dict_filter import filter_list_of_dicts


class NotSupportedError(Exception):
    pass


class IServerError(IOError):
    def __init__(self, message, http_code):
        super().__init__(message)
        self.http_code = http_code


class MstrException(Exception):
    """Base class for exceptions returned by the Strategy One REST API.

    Attributes:
        code: Error code
        message: Error message
        ticket_id: MSTR Ticket ID
    """

    def __init__(self, err_data: dict):
        self.code = err_data.get("code")
        self.message = err_data.get("message")
        self.ticket_id = err_data.get("ticketId")
        self.iserver_code = err_data.get("iServerCode")
        self.full_message = (
            f"{self.code}: {self.message} (Ticket ID: {self.ticket_id}, "
            f"iServerCode: {self.iserver_code})"
        )
        super().__init__(self.full_message)


class MstrTimeoutError(MstrException):
    pass


class VersionException(Exception):
    pass


class IServerException(Exception):
    pass


class PromptedContentError(Exception):
    pass


class Success(Exception):
    """This error holds details about the requested operation.

    Attributes:
        succeeded: list of succeeded operations dict elements
    """

    def __init__(self, data: list[dict]):
        assert isinstance(data, list)

        self.succeeded = data
        self.full_message = (
            f"Operation successful:\n{len(self.succeeded)} succeeded requests"
        )
        super().__init__(self.full_message)

    def __bool__(self):
        return True


class PartialSuccess(Exception):
    """This error holds details about the requested operation.

    Attributes:
        succeeded: list of succeeded operations dict elements
        failed: list of failed operations dict elements
    """

    def __init__(self, data: list[dict]):
        assert isinstance(data, list)

        self.succeeded = filter_list_of_dicts(data, status=[200, 204])
        self.failed = filter_list_of_dicts(data, status=">=400")
        self.full_message = (
            f"Operation partially successful:\n{len(self.failed)} failed "
            f"requests\n{len(self.succeeded)} succeeded requests"
        )
        super().__init__(self.full_message)

    def __bool__(self):
        return False


class Rights(IntFlag):
    """Enumeration constants used to specify the access granted attribute of
    the DSS objects."""

    EXECUTE = 0b10000000  # 128
    USE = 0b01000000  # 64
    CONTROL = 0b00100000  # 32
    DELETE = 0b00010000  # 16
    WRITE = 0b00001000  # 8
    READ = 0b00000100  # 4
    USE_EXECUTE = 0b00000010  # 2 (deprecated)
    BROWSE = 0b00000001  # 1
    INHERITABLE = 0b100000000000000000000000000000  # 536870912


class Permissions(Enum):
    """Enumeration constants used to specify combination of Rights values
    similar to workstation Security Access.

    This has to be string-based to discern between 'Denied All'
    and 'Full Control', which have the same mask.
    """

    DENIED_ALL = 'Denied All'
    DEFAULT_ALL = 'Default All'
    CONSUME = 'Consume'
    VIEW = 'View'
    MODIFY = 'Modify'
    FULL_CONTROL = 'Full Control'


class AggregatedRights(IntFlag):
    """Enumeration constants used to specify combination of Rights values."""

    NONE = 0b00000000  # 0
    CONSUME = 0b01000101  # 69
    VIEW = 0b11000101  # 197
    MODIFY = 0b11011101  # 221
    ALL = 0b11111111  # 255



import logging
from dataclasses import dataclass
from enum import auto
from typing import Any, Optional, Union

from mstrio.connection import Connection
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.format import Color as _Color
from mstrio.utils.helper import Dictable

logger = logging.getLogger(__name__)


@dataclass
class FormatProperty(Dictable):
    """Object that specifies a single format type and it's value

    Attributes:
        type (FormatType): specify what format type is to be set
        value (string): specify what value to set the specified format type to

    FormatType Values:
        Number Category - integer between 0 and 9, default is 9:
            0 - Fixed
            1 - Currency
            2 - Date
            3 - Time
            4 - Percent
            5 - Fraction
            6 - Scientific
            7 - Custom
            8 - Special
            9 - General
        Number Decimal Places - integer between 0 and 30, default is 0
        Number Thousand Separator - boolean (true or false), default is true
        Number Currency Symbol - string of 1-4 characters, default is $
        Number Currency Position - integer between 0 and 3, default is 0:
            0 - at the beginning, for example $10
            1 - at the end, for example 10$
            2 - at the beginning with extra space, for example $ 10
            3 - at the end with extra space, for example 10 $
        Number Format - string of maximum of 255 characters, default value is
            General, example value: ###0.00;[RED]-###0.00
        Number Negative Numbers - integer between 1 and 5, default is 3:
            1 - normal negative format, for example -10
            2 - red color represents negative, -10 would be 10 in red color
            3 - parentheses represent negative, -10 would be (10)
            4 - both red color and parentheses represent negative, -10 would be
                (10) in red color
            5 - minus sign and red color represent negative, -10 would be -10
                in red color
        Alignment Horizontal - integer between 1 and 5, default value is 1:
            1 - General
            2 - Left
            3 - Center
            4 - Right
            5 - Fill
        Alignment Vertical - integer between 1 and 3, default value is 1:
            1 - Top
            2 - Center
            3 - Bottom
        Alignment Text Wrap - boolean (true or false), default is false
        Alignment Text Direction - allowed values are -90, 0 and 90
        Padding Left - double between 0.0 and 100.0, default value is 1.0
        Padding Right - double between 0.0 and 100.0, default value is 1.0
        Padding Top - double between 0.0 and 100.0, default value is 1.0
        Padding Bottom - double between 0.0 and 100.0, default value is 1.0
        Font Name - string representing font, default is Arial, if provided
            font name does not exist or is not supported, default is used
        Font Bold - boolean (true or false), default is false
        Font Italic - boolean (true or false), default is false
        Font Size - integer between 1 and 2147483647, recommended usable range
            is between 8 and 72, other font sizes might render format unreadable
        Font Strikeout - boolean (true or false), default is false
        Font Underline - boolean (true or false), default is false
        Font Color - FormatProperty.Color object representing font color in hex,
            RGB and server readable formats
        Font Script - integer between 1 and 2147483647, default value is 0,
            value represents corresponding font script, if the value has no
            corresponding script, the default will be used
        Border Top Style - integer between 0 and 6, default value is 1:
            0 - empty border
            1 - thin border
            2 - hair border
            3 - dashed border
            4 - dotted border
            5 - thick border
            6 - double border
        Border Right Style - the same rules as Top Style
        Border Left Style - the same rules as Top Style
        Border Bottom Style - the same rules as Top Style
        Border Top Color - FormatProperty.Color object representing border top
         color in hex, RGB and server readable formats
        Border Right Color - FormatProperty.Color object representing border
            right color in hex, RGB and server readable formats
        Border Left Color - FormatProperty.Color object representing border
            left color in hex, RGB and server readable formats
        Border Bottom Color - FormatProperty.Color object representing border
            bottom color in hex, RGB and server readable formats
        Background Fill Style - integer between 0 and 3, default value is 0:
            0 - solid, only has fill color
            1 - transparent
            2 - gradient, has fill color and gradient color
            3 - pattern, has fill color and gradient color
        Background Fill Color - FormatProperty.Color object representing
            background fill color in hex, RGB and server readable formats
        Background Pattern Color - FormatProperty.Color object representing
            background fill color in hex, RGB and server readable formats
        Background Pattern Style - integer between 1 and 18, default value is 1
        Background Gradient Color - FormatProperty.Color object representing
            background gradient color in hex, RGB and server readable formats
        Background Gradient Angle - allowed values are -90, 0 and 90
        Gradient X Offset - allowed values are 0, 50 and 100
        Gradient Y Offset - allowed values are 0, 50 and 100
    """

    class FormatType(AutoName):
        NUMBER_CATEGORY = auto()
        NUMBER_DECIMAL_PLACES = auto()
        NUMBER_THOUSAND_SEPARATOR = auto()
        NUMBER_CURRENCY_SYMBOL = auto()
        NUMBER_CURRENCY_POSITION = auto()
        NUMBER_FORMAT = auto()
        NUMBER_NEGATIVE_NUMBERS = auto()
        ALIGNMENT_HORIZONTAL = auto()
        ALIGNMENT_VERTICAL = auto()
        ALIGNMENT_TEXT_WRAP = auto()
        ALIGNMENT_TEXT_DIRECTION = auto()
        PADDING_LEFT = auto()
        PADDING_RIGHT = auto()
        PADDING_TOP = auto()
        PADDING_BOTTOM = auto()
        FONT_NAME = auto()
        FONT_BOLD = auto()
        FONT_ITALIC = auto()
        FONT_SIZE = auto()
        FONT_STRIKEOUT = auto()
        FONT_UNDERLINE = auto()
        FONT_COLOR = auto()
        FONT_SCRIPT = auto()
        BORDER_TOP_STYLE = auto()
        BORDER_LEFT_STYLE = auto()
        BORDER_BOTTOM_STYLE = auto()
        BORDER_RIGHT_STYLE = auto()
        BORDER_TOP_COLOR = auto()
        BORDER_LEFT_COLOR = auto()
        BORDER_BOTTOM_COLOR = auto()
        BORDER_RIGHT_COLOR = auto()
        BACKGROUND_FILL_COLOR = auto()
        BACKGROUND_PATTERN_COLOR = auto()
        BACKGROUND_PATTERN_STYLE = auto()
        BACKGROUND_FILL_STYLE = auto()
        BACKGROUND_GRADIENT_COLOR = auto()
        BACKGROUND_GRADIENT_ANGLE = auto()
        BACKGROUND_GRADIENT_X_OFFSET = auto()
        BACKGROUND_GRADIENT_Y_OFFSET = auto()
        SERIES_FILL_STYLE = auto()
        SERIES_FILL_COLOR = auto()
        SERIES_GRADIENT_COLOR = auto()
        SERIES_GRADIENT_ANGLE = auto()
        SERIES_GRADIENT_X_OFFSET = auto()
        SERIES_GRADIENT_Y_OFFSET = auto()
        SERIES_PATTERN_STYLE = auto()
        SERIES_PATTERN_COLOR = auto()

    Color = _Color

    _FROM_DICT_MAP = {
        'type': FormatType,
    }
    type: FormatType
    value: Union[str, "FormatProperty.Color"]

    def to_dict(self, camel_case: bool = True) -> dict:
        if isinstance(self.value, FormatProperty.Color):
            return {'type': self.type.value, 'value': self.value.server_value}
        return {'type': self.type.value, 'value': self.value}

    @classmethod
    def from_dict(
        cls,
        source: dict[str, Any],
        connection: Optional["Connection"] = None,
        to_snake_case: bool = True,
    ) -> "FormatProperty":
        if source.get('type'):
            if 'color' in source.get('type'):
                return FormatProperty(
                    type=FormatProperty.FormatType(source.get('type')),
                    value=FormatProperty.Color(server_value=source.get('value')),
                )
            return FormatProperty(
                type=FormatProperty.FormatType(source.get('type')),
                value=source.get('value'),
            )


@dataclass
class MetricFormat(Dictable):
    """Object that specifies the formatting of the metric's values and headers

    Attributes:
        values (list[FormatProperty]): list of format properties for the values
            of the metric
        header (list[FormatProperty]): list of format properties for the header
            of the metric
    """

    _FROM_DICT_MAP = {
        'values': (
            lambda source, connection: [
                FormatProperty.from_dict(content, connection) for content in source
            ]
        ),
        'header': (
            lambda source, connection: [
                FormatProperty.from_dict(content, connection) for content in source
            ]
        ),
    }

    values: list[FormatProperty]
    header: list[FormatProperty]



from typing import TYPE_CHECKING

from mstrio.utils.error_handlers import ErrorHandler

if TYPE_CHECKING:
    from requests import Response

    from mstrio.connection import Connection


@ErrorHandler(
    err_msg="Authentication error. Check user credentials or REST API URL and try again"
)
def login(connection: 'Connection') -> 'Response':
    """Authenticate a user and create an HTTP session on the web server where
    the user's Strategy One sessions are stored.

    This request returns an authorization token (X-MSTR-AuthToken) which will be
    submitted with subsequent requests. The body of the request contains
    the information needed to create the session. The loginMode parameter in
    the body specifies the authentication mode to use. You can authenticate with
    one of the following authentication modes: Standard (1), Anonymous (8),
    LDAP (16) or API Token (4096). Authentication modes can be enabled through
    System Administration REST APIs, if they are supported by the deployment.

    Args:
        connection (Connection): Strategy One REST API connection object

    Returns:
        Complete HTTP response object.
    """

    ENDPOINT = '/api/auth/login'

    auth_data = {
        'loginMode': connection.login_mode,
        'applicationType': connection._application_type,
    }

    if work_set := connection.working_set:
        auth_data['workingSet'] = work_set

    if (max_search := connection.max_search) is not None:
        auth_data['maxSearch'] = max_search

    if app_id := connection.application_id:
        auth_data['applicationId'] = app_id

    if connection.login_mode == 4096:
        auth_data['username'] = connection.api_token
        return connection.post(
            skip_expiration_check=True, endpoint=ENDPOINT, json=auth_data
        )

    auth_data.update(
        {
            'username': connection.username,
            'password': connection._Connection__password,
        }
    )

    return connection.post(
        skip_expiration_check=True, endpoint=ENDPOINT, data=auth_data
    )


@ErrorHandler(err_msg="Failed to logout.")
def logout(connection: 'Connection', error_msg=None, whitelist=None) -> 'Response':
    """Close all existing sessions for the authenticated user.

    Args:
        connection: Strategy One REST API connection object

    Returns:
        Complete HTTP response object.
    """
    return connection.post(
        skip_expiration_check=True,
        endpoint='/api/auth/logout',
        headers={'X-MSTR-ProjectID': None},
    )


def session_renew(connection: 'Connection') -> 'Response':
    """Extends the HTTP and Intelligence Server sessions by resetting the
    timeouts.

    Args:
        connection: Strategy One REST API connection object

    Returns:
        Complete HTTP response object.
    """
    return connection.put(
        skip_expiration_check=True,
        endpoint='/api/sessions',
        headers={'X-MSTR-ProjectID': None},
        timeout=2.0,
    )


def session_status(connection: 'Connection') -> 'Response':
    """Checks Intelligence Server session status.

    Args:
        connection: Strategy One REST API connection object

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        skip_expiration_check=True,
        endpoint='/api/sessions',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Could not get identity token.")
def identity_token(connection: 'Connection') -> 'Response':
    """Create a new identity token.

    An identity token is used to share an existing session with another
    project, based on the authorization token for the existing
    session.

    Args:
        connection: Strategy One REST API connection object

    Returns:
        Complete HTTP response object.
    """
    return connection.post(
        endpoint='/api/auth/identityToken',
    )


@ErrorHandler(err_msg="Could not get API token.")
def api_token(
    connection: 'Connection', target_user_id: str | None = None
) -> 'Response':
    """Create a new API token.

    An API token is used to authenticate a user via login mode 4096.

    Args:
        connection: Strategy One REST API connection object.
        target_user_id: User id of the user for whom the API token
            is to be created. When omitted, token is for `connection`'s user.

    Returns:
        Complete HTTP response object.
    """
    kwargs = {}
    if target_user_id:
        kwargs['json'] = {'userId': target_user_id}

    return connection.post(
        endpoint='/api/auth/apiTokens',
        **kwargs,
    )


def validate_identity_token(
    connection: 'Connection', identity_token: str
) -> 'Response':
    """Validate an identity token.

    Args:
        connection: Strategy One REST API connection object
        identity_token: Identity token

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint='/api/auth/identityToken',
        headers={'X-MSTR-IdentityToken': identity_token},
    )


@ErrorHandler(
    err_msg="Error creating a new Web server session that shares an existing IServer "
    "session."
)
def delegate(
    connection: 'Connection', identity_token: str, whitelist=None
) -> 'Response':
    """Returns authentication token and cookies from given X-MSTR-
    IdentityToken.

    Args:
        connection: Strategy One REST API connection object
        identity_token: Identity token
        whitelist: list of errors for which we skip printing error messages

    Returns:
        Complete HTTP response object.
    """
    return connection.post(
        skip_expiration_check=True,
        endpoint='/api/auth/delegate',
        json={'loginMode': "-1", 'identityToken': identity_token},
    )


@ErrorHandler(err_msg="Error getting privileges list.")
def user_privileges(connection: 'Connection') -> 'Response':
    """Get the list of privileges for the authenticated user.

    The response includes the name, ID, and description of each
    privilege and specifies which projects the privileges are valid for.

    Args:
        connection: Strategy One REST API connection object

    Returns:
        Complete HTTP response object.
    """
    return connection.get(endpoint='/api/sessions/privileges')


@ErrorHandler(err_msg="Error getting info for authenticated user.")
def get_info_for_authenticated_user(
    connection: 'Connection', error_msg=None
) -> 'Response':
    """Get information for the authenticated user.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(endpoint='/api/sessions/userInfo')



import datetime as dt
from base64 import b64encode

from pandas import DataFrame


class Encoder:
    """Internal method for converting a Pandas DataFrame to Strategy One
    compliant base64 encoded JSON.

    When creating a data set, Strategy One APIs require the tabular data to
    have been transformed first into JSON and then into a base 64 encoded string
    before it is transmitted to the Intelligence Server via the REST API to
    create the data set. This class uses Pandas to handle transforming
    the DataFrame into a JSON representation of the data. For single-table data
    sets, Strategy One APIs require the JSON data to be formatted using
    the 'records' orientation from Pandas. Conversely, multi-table data sets
    require the JSON data to have a 'values' orientation. Based on the data set
    type, the correct encoding strategy is applied and the data is then encoded.

    Attributes:
        data_frame: Pandas DataFrame to be encoded.
        b64_data: DataFrame converted into a Base-64 encoded JSON string.
        orientation: For single-table data sets, "single"; for multi-table
            data sets, "multi".
    """

    def __init__(self, data_frame: DataFrame, dataset_type: str):
        """Inits Encoder with given data_frame and type.

        Args:
            data_frame (DataFrame): Pandas DataFrame to be converted.
            dataset_type (str): Dataset type. One of `single` or `multi` to
                correspond with single-table or multi-table sources.
        """
        self.data_frame: DataFrame = data_frame
        self._b64_data: str | None = None
        self.orientation: str | None = None
        # Mapping used when converting DataFrame rows
        # into proper JSON orientation needed for data uploads.
        _table_type_orient_map: dict = {'single': 'records', 'multi': 'values'}

        # Sets the proper orientation
        if dataset_type not in _table_type_orient_map:
            allowed_types = ', '.join(_table_type_orient_map)
            raise ValueError(f"Table type should be one of {allowed_types}")

        self.orientation = _table_type_orient_map[dataset_type]

    def encode(self) -> None:
        """Encode data in base 64."""
        self.data_frame = self.data_frame.apply(
            lambda x: x.astype('str') if isinstance(x.iloc[0], dt.date) else x
        )

        json_data = self.data_frame.to_json(orient=self.orientation, date_format='iso')
        self._b64_data = b64encode(json_data.encode('utf-8')).decode('utf-8')

    @property
    def b64_data(self) -> str:
        if not self._b64_data:
            self.encode()
        return self._b64_data
